"""
Training utilities for PyTorch Lightning
"""
import os
import json
from typing import List, Dict, Any, Optional, Union, Tuple, Callable

import lightning as L
from lightning.pytorch.callbacks import (
    EarlyStopping, 
    ModelCheckpoint, 
    LearningRateMonitor
)
from lightning.pytorch.loggers import TensorBoardLogger
import torch
from torch.utils.data import DataLoader, Dataset

from .callbacks import ArchitectureLogger, ArchitectureVisualizer


class LightningTrainer:
    """
    Wrapper for PyTorch Lightning Trainer with additional functionality
    
    Args:
        max_epochs (int, optional): Maximum number of epochs. Defaults to 100.
        patience (int, optional): Early stopping patience. Defaults to 10.
        monitor (str, optional): Metric to monitor. Defaults to "val_loss".
        mode (str, optional): Optimization mode. Defaults to "min".
        accelerator (str, optional): Accelerator type. Defaults to "auto".
        devices (Union[int, str, List[int]], optional): Devices to use. Defaults to "auto".
        checkpoint_dir (Optional[str], optional): Directory for checkpoints. Defaults to None.
        log_dir (Optional[str], optional): Directory for logs. Defaults to None.
    """
    
    def __init__(
        self,
        max_epochs: int = 100,
        patience: int = 10,
        monitor: str = "val_loss",
        mode: str = "min",
        accelerator: str = "auto",
        devices: Union[int, str, List[int]] = "auto",
        checkpoint_dir: Optional[str] = None,
        log_dir: Optional[str] = None
    ):
        self.max_epochs = max_epochs
        self.patience = patience
        self.monitor = monitor
        self.mode = mode
        self.accelerator = accelerator
        self.devices = devices
        self.checkpoint_dir = checkpoint_dir or "./checkpoints"
        self.log_dir = log_dir or "./logs"
        
        os.makedirs(self.checkpoint_dir, exist_ok=True)
        os.makedirs(self.log_dir, exist_ok=True)
        
        self.trainer = None
        self.callbacks = []
    
    def setup_trainer(
        self, 
        additional_callbacks: Optional[List[Callable]] = None,
        logger: bool = True,
        enable_progress_bar: bool = True,
        **trainer_kwargs
    ) -> L.Trainer:
        """
        Set up the Lightning Trainer
        
        Args:
            additional_callbacks (Optional[List[Callable]], optional): Additional callbacks. Defaults to None.
            logger (bool, optional): Whether to use logger. Defaults to True.
            enable_progress_bar (bool, optional): Whether to enable progress bar. Defaults to True.
            **trainer_kwargs: Additional trainer arguments
            
        Returns:
            L.Trainer: Lightning Trainer
        """
        # Basic callbacks
        self.callbacks = [
            EarlyStopping(
                monitor=self.monitor,
                patience=self.patience,
                mode=self.mode,
                verbose=True
            ),
            ModelCheckpoint(
                dirpath=self.checkpoint_dir,
                filename=f"best-{{epoch}}-{{{self.monitor}:.4f}}",
                monitor=self.monitor,
                mode=self.mode,
                save_top_k=1,
                verbose=True
            ),
            LearningRateMonitor(logging_interval="epoch"),
            ArchitectureLogger(log_dir=os.path.join(self.log_dir, "architecture")),
            ArchitectureVisualizer(log_dir=os.path.join(self.log_dir, "visualizations"))
        ]
        
        # Add additional callbacks
        if additional_callbacks:
            self.callbacks.extend(additional_callbacks)
        
        # Logger
        tb_logger = TensorBoardLogger(save_dir=self.log_dir) if logger else False
        
        # Create trainer
        self.trainer = L.Trainer(
            max_epochs=self.max_epochs,
            callbacks=self.callbacks,
            accelerator=self.accelerator,
            devices=self.devices,
            logger=tb_logger,
            enable_progress_bar=enable_progress_bar,
            **trainer_kwargs
        )
        
        return self.trainer
    
    def fit(
        self, 
        model: L.LightningModule, 
        train_dataloader: DataLoader, 
        val_dataloader: Optional[DataLoader] = None,
        **trainer_kwargs
    ) -> L.Trainer:
        """
        Train the model
        
        Args:
            model (L.LightningModule): Lightning model
            train_dataloader (DataLoader): Training dataloader
            val_dataloader (Optional[DataLoader], optional): Validation dataloader. Defaults to None.
            **trainer_kwargs: Additional trainer arguments
            
        Returns:
            L.Trainer: Lightning Trainer
        """
        # Set up trainer if not already set up
        if self.trainer is None:
            self.setup_trainer(**trainer_kwargs)
        
        # Train model
        self.trainer.fit(model, train_dataloader, val_dataloader)
        
        return self.trainer
    
    def test(
        self, 
        model: L.LightningModule, 
        test_dataloader: DataLoader
    ) -> Dict[str, float]:
        """
        Test the model
        
        Args:
            model (L.LightningModule): Lightning model
            test_dataloader (DataLoader): Test dataloader
            
        Returns:
            Dict[str, float]: Test metrics
        """
        if self.trainer is None:
            raise ValueError("Trainer not set up yet. Call setup_trainer() or fit() first.")
        
        # Test model
        results = self.trainer.test(model, test_dataloader)
        
        # Save results
        results_file = os.path.join(self.log_dir, "test_results.json")
        with open(results_file, "w") as f:
            json.dump(results[0], f, indent=2)
        
        return results[0]
    
    def save_model_summary(self, model: L.LightningModule, filename: str = "model_summary.txt") -> None:
        """
        Save model summary to file
        
        Args:
            model (L.LightningModule): Lightning model
            filename (str, optional): Output filename. Defaults to "model_summary.txt".
        """
        # Create summary
        summary = str(model)
        
        # Save to file
        summary_file = os.path.join(self.log_dir, filename)
        with open(summary_file, "w") as f:
            f.write(summary)
    
    def get_best_model_path(self) -> Optional[str]:
        """
        Get path to the best model checkpoint
        
        Returns:
            Optional[str]: Path to best model checkpoint
        """
        if self.trainer is None:
            raise ValueError("Trainer not set up yet. Call setup_trainer() or fit() first.")
        
        # Get checkpoint callback
        checkpoint_callback = None
        for callback in self.callbacks:
            if isinstance(callback, ModelCheckpoint):
                checkpoint_callback = callback
                break
        
        if checkpoint_callback is None:
            return None
        
        return checkpoint_callback.best_model_path
    
    def load_best_model(self, model_class: type, **model_kwargs) -> L.LightningModule:
        """
        Load the best model from checkpoint
        
        Args:
            model_class (type): Lightning model class
            **model_kwargs: Model initialization arguments
            
        Returns:
            L.LightningModule: Loaded model
        """
        best_model_path = self.get_best_model_path()
        
        if best_model_path is None or not os.path.exists(best_model_path):
            raise ValueError("No best model checkpoint found.")
        
        # Load model
        model = model_class.load_from_checkpoint(best_model_path, **model_kwargs)
        
        return model


class DataModule(L.LightningDataModule):
    """
    Lightning DataModule for standardized data handling
    
    Args:
        train_dataset (Dataset): Training dataset
        val_dataset (Optional[Dataset], optional): Validation dataset. Defaults to None.
        test_dataset (Optional[Dataset], optional): Test dataset. Defaults to None.
        batch_size (int, optional): Batch size. Defaults to 32.
        num_workers (int, optional): Number of worker processes. Defaults to 4.
        shuffle (bool, optional): Whether to shuffle training data. Defaults to True.
        pin_memory (bool, optional): Whether to pin memory. Defaults to True.
        drop_last (bool, optional): Whether to drop last incomplete batch. Defaults to False.
    """
    
    def __init__(
        self,
        train_dataset: Dataset,
        val_dataset: Optional[Dataset] = None,
        test_dataset: Optional[Dataset] = None,
        batch_size: int = 32,
        num_workers: int = 4,
        shuffle: bool = True,
        pin_memory: bool = True,
        drop_last: bool = False
    ):
        super().__init__()
        self.train_dataset = train_dataset
        self.val_dataset = val_dataset
        self.test_dataset = test_dataset
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.shuffle = shuffle
        self.pin_memory = pin_memory
        self.drop_last = drop_last
    
    def train_dataloader(self) -> DataLoader:
        """Get training dataloader"""
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=self.shuffle,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            drop_last=self.drop_last
        )
    
    def val_dataloader(self) -> Optional[DataLoader]:
        """Get validation dataloader"""
        if self.val_dataset is None:
            return None
        
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory
        )
    
    def test_dataloader(self) -> Optional[DataLoader]:
        """Get test dataloader"""
        if self.test_dataset is None:
            return None
        
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory
        )



"""
PyDeevo utilities module initialization
"""
from .helpers import (
    setup_logging,
    get_device,
    save_dict_to_json,
    load_dict_from_json,
    plot_search_progress,
    visualize_architecture_comparison,
    calculate_model_complexity
)

from .profiling import (
    ModelProfiler,
    ArchitectureBenchmarker,
    LightningProfileCallback,
    FlopsCalculator
)

from .distributed import (
    DistributedTrainingHelper,
    MemoryOptimization,
    ModelShardingHelper,
    BatchSizeOptimizer,
    SecurityWrapper
)

from .export import (
    ModelExporter,
    InferenceProfiler,
    InferenceOptimizer
)

from .data import (
    PolarsDataProcessor,
    PolarsDataset,
    DatasetBuilder,
    handle_missing_values,
    encode_categorical,
    normalize_features,
    create_polynomial_features
)

from .database import (
    DuckDBManager,
    SQLDataProcessor,
    AnalyticalDataManager,
    DuckDBDataset,
    sql_to_pytorch_dataset,
    execute_analytical_query,
    etl_pipeline
)

__all__ = [
    # From helpers
    'setup_logging',
    'get_device',
    'save_dict_to_json',
    'load_dict_from_json',
    'plot_search_progress',
    'visualize_architecture_comparison',
    'calculate_model_complexity',
    
    # From profiling
    'ModelProfiler',
    'ArchitectureBenchmarker',
    'LightningProfileCallback',
    'FlopsCalculator',
    
    # From distributed
    'DistributedTrainingHelper',
    'MemoryOptimization',
    'ModelShardingHelper',
    'BatchSizeOptimizer',
    'SecurityWrapper',
    
    # From export
    'ModelExporter',
    'InferenceProfiler',
    'InferenceOptimizer',
    
    # From data
    'PolarsDataProcessor',
    'PolarsDataset',
    'DatasetBuilder',
    'handle_missing_values',
    'encode_categorical',
    'normalize_features',
    'create_polynomial_features',
    
    # From database
    'DuckDBManager',
    'SQLDataProcessor',
    'AnalyticalDataManager',
    'DuckDBDataset',
    'sql_to_pytorch_dataset',
    'execute_analytical_query',
    'etl_pipeline'
]



"""
Example using DuckDB with PyDeevo for SQL-based data analytics
"""
import os
import polars as pl
import torch
from torch.utils.data import DataLoader
import lightning as L
import matplotlib.pyplot as plt
import numpy as np

from pydeevo.models.base import FlexibleModule
from pydeevo.utils.database import (
    DuckDBManager,
    AnalyticalDataManager,
    sql_to_pytorch_dataset,
    execute_analytical_query
)


def main():
    """Example of using DuckDB with PyDeevo"""
    print("Setting up analytics environment with DuckDB...")
    
    # Set up directories
    os.makedirs("./duckdb_example", exist_ok=True)
    db_path = "./duckdb_example/analytics.duckdb"
    
    # Create sample data using Polars
    print("Creating sample data...")
    # Customer data
    customers = pl.DataFrame({
        "customer_id": list(range(1, 1001)),
        "age": np.random.randint(18, 80, 1000),
        "income": np.random.lognormal(10, 1, 1000),
        "segment": np.random.choice(["A", "B", "C"], 1000)
    })
    
    # Transaction data
    transactions = pl.DataFrame({
        "transaction_id": list(range(1, 10001)),
        "customer_id": np.random.randint(1, 1001, 10000),
        "amount": np.random.lognormal(4, 1, 10000),
        "category": np.random.choice(["Food", "Electronics", "Clothing", "Services"], 10000),
        "timestamp": np.random.randint(1600000000, 1700000000, 10000)
    })
    
    # Set up DuckDB and load data
    print("Loading data into DuckDB...")
    analytics = AnalyticalDataManager(db_path, cache_dir="./duckdb_example/cache")
    analytics.load_data(customers, "customers")
    analytics.load_data(transactions, "transactions")
    
    # Run some analytical queries
    print("Running analytical queries...")
    
    # Customer spending by segment
    segment_query = """
    SELECT 
        c.segment, 
        COUNT(DISTINCT c.customer_id) as customer_count,
        COUNT(t.transaction_id) as transaction_count,
        SUM(t.amount) as total_spend,
        AVG(t.amount) as avg_transaction,
        SUM(t.amount) / COUNT(DISTINCT c.customer_id) as spend_per_customer
    FROM customers c
    JOIN transactions t ON c.customer_id = t.customer_id
    GROUP BY c.segment
    ORDER BY total_spend DESC
    """
    
    segment_results = analytics.execute_sql(
        segment_query, 
        cache_key="segment_analysis"
    )
    
    print("\nCustomer Spending by Segment:")
    print(segment_results)
    
    # Category popularity
    category_query = """
    SELECT 
        t.category,
        COUNT(*) as transaction_count,
        SUM(t.amount) as total_spend,
        AVG(t.amount) as avg_amount
    FROM transactions t
    GROUP BY t.category
    ORDER BY total_spend DESC
    """
    
    category_results = analytics.execute_sql(
        category_query, 
        cache_key="category_analysis"
    )
    
    print("\nCategory Popularity:")
    print(category_results)
    
    # High-value customers
    top_customers_query = """
    SELECT 
        c.customer_id,
        c.age,
        c.segment,
        COUNT(t.transaction_id) as transaction_count,
        SUM(t.amount) as total_spend
    FROM customers c
    JOIN transactions t ON c.customer_id = t.customer_id
    GROUP BY c.customer_id, c.age, c.segment
    ORDER BY total_spend DESC
    LIMIT 20
    """
    
    top_customers = analytics.execute_sql(
        top_customers_query, 
        cache_key="top_customers"
    )
    
    print("\nTop 5 Customers by Spend:")
    print(top_customers.head(5))
    
    # Prepare data for machine learning model
    print("\nPreparing data for machine learning...")
    
    # Get feature data for predicting customer spending
    feature_query = """
    SELECT 
        c.customer_id,
        c.age,
        CASE WHEN c.segment = 'A' THEN 1 WHEN c.segment = 'B' THEN 2 ELSE 3 END as segment_num,
        c.income,
        COUNT(t.transaction_id) as transaction_count,
        AVG(t.amount) as avg_transaction,
        SUM(t.amount) as total_spend
    FROM customers c
    JOIN transactions t ON c.customer_id = t.customer_id
    GROUP BY c.customer_id, c.age, c.segment, c.income
    """
    
    # Create a PyTorch dataset directly from SQL
    dataset = sql_to_pytorch_dataset(
        db_path=db_path,
        query=feature_query,
        feature_cols=["age", "segment_num", "income", "transaction_count", "avg_transaction"],
        target_col="total_spend",
        batch_size=100
    )
    
    print(f"Created dataset with {len(dataset)} samples")
    
    # Split into train, validation, test sets
    train_size = int(0.7 * len(dataset))
    val_size = int(0.15 * len(dataset))
    test_size = len(dataset) - train_size - val_size
    
    train_dataset, val_dataset, test_dataset = torch.utils.data.random_split(
        dataset, [train_size, val_size, test_size]
    )
    
    # Create data loaders
    train_loader = DataLoader(train_dataset, batch_size=32, shuffle=True)
    val_loader = DataLoader(val_dataset, batch_size=32)
    test_loader = DataLoader(test_dataset, batch_size=32)
    
    # Create and train a regression model
    print("Creating and training model...")
    
    model = FlexibleModule(
        architecture=[5, 64, 32, 16, 1],  # 5 features, 1 output
        learning_rate=0.001,
        loss_fn=torch.nn.MSELoss()
    )
    
    trainer = L.Trainer(
        max_epochs=50,
        enable_progress_bar=True,
        log_every_n_steps=10,
        enable_model_summary=True,
    )
    
    trainer.fit(model, train_loader, val_loader)
    
    # Test model
    results = trainer.test(model, test_loader)
    print(f"Test results: {results}")
    
    # Make predictions for sample data
    print("Making predictions...")
    model.eval()
    predictions = []
    actuals = []
    
    with torch.no_grad():
        for batch in test_loader:
            features, target = batch
            output = model(features)
            predictions.extend(output.squeeze().cpu().numpy())
            actuals.extend(target.squeeze().cpu().numpy())
    
    # Plot predictions vs actuals
    plt.figure(figsize=(10, 6))
    plt.scatter(actuals, predictions, alpha=0.5)
    
    # Add perfect prediction line
    min_val = min(min(actuals), min(predictions))
    max_val = max(max(actuals), max(predictions))
    plt.plot([min_val, max_val], [min_val, max_val], 'r--')
    
    plt.xlabel('Actual Spending')
    plt.ylabel('Predicted Spending')
    plt.title('Customer Spending Prediction')
    plt.grid(True)
    
    plt.savefig("./duckdb_example/predictions.png")
    print("Results saved to ./duckdb_example/predictions.png")
    
    # Run an analytical pipeline
    print("\nRunning analytical pipeline...")
    
    pipeline = [
        {
            "type": "sql",
            "name": "age_brackets",
            "query": """
            SELECT 
                CASE 
                    WHEN age < 30 THEN 'Under 30'
                    WHEN age >= 30 AND age < 45 THEN '30-44'
                    WHEN age >= 45 AND age < 60 THEN '45-59'
                    ELSE '60+' 
                END as age_group,
                COUNT(*) as customer_count,
                AVG(income) as avg_income
            FROM customers
            GROUP BY age_group
            ORDER BY age_group
            """,
            "cache_key": "age_brackets"
        },
        {
            "type": "sql",
            "name": "age_spending",
            "query": """
            SELECT 
                CASE 
                    WHEN c.age < 30 THEN 'Under 30'
                    WHEN c.age >= 30 AND c.age < 45 THEN '30-44'
                    WHEN c.age >= 45 AND c.age < 60 THEN '45-59'
                    ELSE '60+' 
                END as age_group,
                AVG(t.amount) as avg_transaction,
                SUM(t.amount) / COUNT(DISTINCT c.customer_id) as spend_per_customer
            FROM customers c
            JOIN transactions t ON c.customer_id = t.customer_id
            GROUP BY age_group
            ORDER BY age_group
            """,
            "cache_key": "age_spending"
        },
        {
            "type": "transform",
            "name": "combined_age_analysis",
            "input": "age_brackets",
            "transform_type": "join",
            "right": "age_spending",
            "on": "age_group",
            "how": "inner"
        },
        {
            "type": "export",
            "input": "combined_age_analysis",
            "path": "./duckdb_example/age_analysis.csv",
            "export_type": "csv"
        }
    ]
    
    pipeline_results = analytics.run_analytical_pipeline(pipeline)
    
    print("\nAge Group Analysis:")
    print(pipeline_results["combined_age_analysis"])
    print("\nAnalysis exported to ./duckdb_example/age_analysis.csv")
    
    print("\nDuckDB example completed!")


if __name__ == "__main__":
    main()



, filename))
        plt.close()
    
    def _visualize_cnn(self, architecture: Tuple[List[Dict[str, Any]], List[int]], filename: str) -> None:
        """Visualize CNN architecture"""
        conv_layers, fc_layers = architecture
        
        # Create figure with two subplots - one for the overall architecture, one for details
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(12, 10), gridspec_kw={'height_ratios': [1, 2]})
        
        # Top subplot: Overall architecture diagram
        layer_types = ['Input'] + ['Conv'] * len(conv_layers) + ['FC'] * len(fc_layers)
        layer_names = ['Input']
        
        # Add conv layer names
        for i, layer in enumerate(conv_layers):
            filters = layer.get('filters', 0)
            kernel = layer.get('kernel_size', 3)
            layer_names.append(f"Conv\n{filters}@{kernel}x{kernel}")
        
        # Add fc layer names
        for neurons in fc_layers:
            layer_names.append(f"FC\n{neurons}")
        
        # Draw boxes for each layer
        n_layers = len(layer_names)
        box_width = 0.8
        box_spacing = 1.2
        
        for i, (name, layer_type) in enumerate(zip(layer_names, layer_types)):
            # Position
            x = i * box_spacing
            y = 0
            
            # Color based on layer type
            if layer_type == 'Input':
                color = 'lightgreen'
            elif layer_type == 'Conv':
                color = 'lightblue'
            else:  # FC
                color = 'lightcoral'
            
            # Draw box
            rect = plt.Rectangle((x - box_width/2, y - 0.5), box_width, 1, 
                                 fill=True, color=color, zorder=3)
            ax1.add_patch(rect)
            
            # Add text
            ax1.text(x, y, name, ha='center', va='center', zorder=4)
            
            # Draw arrows
            if i < n_layers - 1:
                ax1.arrow(x + box_width/2, y, box_spacing - box_width, 0, 
                         head_width=0.1, head_length=0.1, fc='black', ec='black', zorder=2)
        
        # Set limits and turn off axis
        ax1.set_xlim(-1, (n_layers - 1) * box_spacing + 1)
        ax1.set_ylim(-1, 1)
        ax1.axis('off')
        ax1.set_title("CNN Architecture Overview")
        
        # Bottom subplot: Detailed architecture information
        # Create table data
        table_data = [
            ["Layer", "Type", "Output Shape", "Parameters"]
        ]
        
        # Add input layer
        input_shape = (3, 32, 32)  # Placeholder, should be replaced with actual shape
        table_data.append(["Input", "Input", str(input_shape), "0"])
        
        # Track current shape
        c, h, w = input_shape
        param_count = 0
        
        # Add conv layers
        for i, layer in enumerate(conv_layers):
            filters = layer.get('filters', 0)
            kernel = layer.get('kernel_size', 3)
            padding = layer.get('padding', 0)
            stride = layer.get('stride', 1)
            pool_size = layer.get('pool_size', 1)
            
            # Calculate output shape after conv
            h_out = (h - kernel + 2 * padding) // stride + 1
            w_out = (w - kernel + 2 * padding) // stride + 1
            
            # Calculate parameters
            params = filters * (c * kernel * kernel + 1)  # +1 for bias
            param_count += params
            
            # Add to table
            output_shape = f"({filters}, {h_out}, {w_out})"
            table_data.append([f"Conv{i+1}", f"{kernel}x{kernel} Conv", output_shape, f"{params:,}"])
            
            # Update shape for next layer
            c, h, w = filters, h_out, w_out
            
            # Add pooling if applicable
            if pool_size > 1:
                h_out = h // pool_size
                w_out = w // pool_size
                
                output_shape = f"({c}, {h_out}, {w_out})"
                table_data.append([f"Pool{i+1}", f"{pool_size}x{pool_size} MaxPool", output_shape, "0"])
                
                # Update shape for next layer
                h, w = h_out, w_out
        
        # Add flatten layer
        flattened_size = c * h * w
        table_data.append(["Flatten", "Flatten", f"({flattened_size})", "0"])
        
        # Add fully connected layers
        fc_input_size = flattened_size
        for i, neurons in enumerate(fc_layers):
            # Calculate parameters
            params = fc_input_size * neurons + neurons  # weights + biases
            param_count += params
            
            # Add to table
            table_data.append([f"FC{i+1}", "Fully Connected", f"({neurons})", f"{params:,}"])
            
            # Update shape for next layer
            fc_input_size = neurons
        
        # Add total row
        table_data.append(["Total", "", "", f"{param_count:,}"])
        
        # Create table
        ax2.axis('tight')
        ax2.axis('off')
        
        table = ax2.table(cellText=table_data[1:], 
                         colLabels=table_data[0], 
                         loc='center',
                         cellLoc='center')
        
        table.auto_set_font_size(False)
        table.set_fontsize(10)
        table.scale(1, 1.5)
        
        ax2.set_title("CNN Architecture Details")
        
        # Adjust layout and save
        plt.tight_layout()
        plt.savefig(os.path.join(self.log_dir, filename))
        plt.close()



"""
PyDeevo: Evolutionary Neural Architecture Search with Hyperparameter Optimization
===============================================================================

A modular framework integrating PyTorch Lightning, Optuna, and PyGAD for multi-level
optimization of neural network architectures and hyperparameters.
"""

__version__ = "0.1.0"



"""
Advanced utilities for distributed training and memory optimization
"""
import os
import logging
import json
import time
from typing import Dict, List, Any, Optional, Union, Tuple, Callable

import torch
import lightning as L
from lightning.fabric import Fabric
from lightning.fabric.strategies import FSDPStrategy, DDPStrategy
from torch.distributed.fsdp import (
    FullyShardedDataParallel as FSDP,
    CPUOffload,
    MixedPrecision,
    BackwardPrefetch,
    ShardingStrategy,
    StateDictType,
)
from torch.distributed.fsdp.wrap import transformer_auto_wrap_policy
from safetensors.torch import save_file, load_file

logger = logging.getLogger(__name__)


class DistributedTrainingHelper:
    """
    Helper class for distributed training with Lightning Fabric
    
    This class provides utilities for distributed training with Lightning Fabric,
    including FSDP (Fully Sharded Data Parallel) support for training large models.
    
    Args:
        model_size_gb (float, optional): Estimated model size in GB. Defaults to None.
        precision (str, optional): Precision to use for training. Defaults to '16-mixed'.
        devices (Union[int, str, List[int]], optional): Devices to use. Defaults to 'auto'.
        strategy (str, optional): Distributed strategy to use. Defaults to 'auto'.
        checkpoint_dir (Optional[str], optional): Directory for checkpoints. Defaults to None.
        cpu_offload (bool, optional): Whether to offload parameters to CPU. Defaults to False.
        mixed_precision (bool, optional): Whether to use mixed precision. Defaults to True.
        activation_checkpointing (bool, optional): Whether to use activation checkpointing. Defaults to False.
        profiling (bool, optional): Whether to enable profiling. Defaults to False.
    """
    
    def __init__(
        self,
        model_size_gb: Optional[float] = None,
        precision: str = '16-mixed',
        devices: Union[int, str, List[int]] = 'auto',
        strategy: str = 'auto',
        checkpoint_dir: Optional[str] = None,
        cpu_offload: bool = False,
        mixed_precision: bool = True,
        activation_checkpointing: bool = False,
        profiling: bool = False,
    ):
        self.model_size_gb = model_size_gb
        self.precision = precision
        self.devices = devices
        self.strategy = strategy
        self.checkpoint_dir = checkpoint_dir or "./checkpoints"
        self.cpu_offload = cpu_offload
        self.mixed_precision = mixed_precision
        self.activation_checkpointing = activation_checkpointing
        self.profiling = profiling
        
        os.makedirs(self.checkpoint_dir, exist_ok=True)
        
        self.fabric = None
        self.strategy_obj = None
        self.profiler = None
    
    def setup_fabric(self, transformer_modules: Optional[List[type]] = None):
        """
        Set up Lightning Fabric for distributed training
        
        Args:
            transformer_modules (Optional[List[type]], optional): List of transformer module types
                for auto wrapping policy. Defaults to None.
                
        Returns:
            lightning.fabric.Fabric: Configured Fabric instance
        """
        # Determine optimal strategy based on model size and available resources
        if self.strategy == 'auto':
            self.strategy = self._select_optimal_strategy()
        
        # Create appropriate strategy object
        if self.strategy == 'fsdp':
            # Configure FSDP strategy
            mp_config = None
            if self.mixed_precision:
                mp_config = MixedPrecision(
                    param_dtype=torch.float16,
                    reduce_dtype=torch.float16,
                    buffer_dtype=torch.float16,
                )
            
            # Configure auto wrapping policy if transformer modules provided
            auto_wrap_policy = None
            if transformer_modules:
                auto_wrap_policy = transformer_auto_wrap_policy(
                    transformer_layer_cls=tuple(transformer_modules)
                )
            
            # Create FSDP strategy
            self.strategy_obj = FSDPStrategy(
                auto_wrap_policy=auto_wrap_policy,
                cpu_offload=CPUOffload(offload_params=self.cpu_offload),
                mixed_precision=mp_config,
                backward_prefetch=BackwardPrefetch.BACKWARD_PRE,
                activation_checkpointing=self.activation_checkpointing,
                sharding_strategy=ShardingStrategy.FULL_SHARD,
                state_dict_type=StateDictType.FULL_STATE_DICT,
                limit_all_gathers=True,
            )
        elif self.strategy == 'ddp':
            # Create DDP strategy
            self.strategy_obj = DDPStrategy(
                find_unused_parameters=False,
                gradient_as_bucket_view=True,
            )
        
        # Set up profiler if enabled
        if self.profiling:
            from lightning.pytorch.profilers import PyTorchProfiler
            self.profiler = PyTorchProfiler(
                activities=[
                    torch.profiler.ProfilerActivity.CPU,
                    torch.profiler.ProfilerActivity.CUDA,
                ],
                schedule=torch.profiler.schedule(
                    wait=1,
                    warmup=1,
                    active=3,
                ),
                on_trace_ready=torch.profiler.tensorboard_trace_handler(
                    os.path.join(self.checkpoint_dir, "profiler_traces")
                ),
                record_shapes=True,
                profile_memory=True,
            )
        
        # Create Fabric instance
        self.fabric = Fabric(
            accelerator="auto",
            devices=self.devices,
            precision=self.precision,
            strategy=self.strategy_obj,
            loggers=None,  # Can be configured separately
            callbacks=None,  # Can be configured separately
            profiler=self.profiler,
        )
        
        return self.fabric
    
    def _select_optimal_strategy(self) -> str:
        """
        Select the optimal distributed strategy based on model size and available resources
        
        Returns:
            str: Selected strategy ('fsdp', 'ddp', or 'single')
        """
        # Check if CUDA is available
        if not torch.cuda.is_available():
            return 'single'
        
        # Get number of GPUs
        num_gpus = torch.cuda.device_count()
        
        if num_gpus <= 1:
            return 'single'
        
        # If model size is provided, use it to determine strategy
        if self.model_size_gb is not None:
            # Calculate available GPU memory
            gpu_mem_gb = torch.cuda.get_device_properties(0).total_memory / (1024**3)
            
            # If model doesn't fit in a single GPU, use FSDP
            if self.model_size_gb > gpu_mem_gb * 0.8:  # 80% of GPU memory
                return 'fsdp'
        
        # Default to DDP for multi-GPU setups
        return 'ddp' if num_gpus > 1 else 'single'
    
    def setup_model_and_optimizer(
        self, 
        model: torch.nn.Module, 
        optimizer_fn: Callable,
        setup_module_fn: Optional[Callable] = None
    ) -> Tuple[torch.nn.Module, torch.optim.Optimizer]:
        """
        Set up model and optimizer with Fabric
        
        Args:
            model (torch.nn.Module): PyTorch model
            optimizer_fn (Callable): Function to create optimizer given model parameters
            setup_module_fn (Optional[Callable], optional): Function to set up model modules
                before wrapping with Fabric. Defaults to None.
                
        Returns:
            Tuple[torch.nn.Module, torch.optim.Optimizer]: Configured model and optimizer
        """
        if self.fabric is None:
            raise ValueError("Fabric not set up. Call setup_fabric() first.")
        
        # Apply custom module setup if provided
        if setup_module_fn is not None:
            model = setup_module_fn(model)
        
        # Create optimizer
        optimizer = optimizer_fn(model.parameters())
        
        # Set up model and optimizer with Fabric
        model, optimizer = self.fabric.setup(model, optimizer)
        
        return model, optimizer
    
    def save_model_checkpoint(
        self, 
        model: torch.nn.Module, 
        optimizer: torch.optim.Optimizer, 
        epoch: int, 
        metrics: Dict[str, Any],
        filename: Optional[str] = None,
        save_with_safetensors: bool = True,
    ) -> str:
        """
        Save model checkpoint with Fabric
        
        Args:
            model (torch.nn.Module): PyTorch model
            optimizer (torch.optim.Optimizer): Optimizer
            epoch (int): Current epoch
            metrics (Dict[str, Any]): Training metrics
            filename (Optional[str], optional): Checkpoint filename. Defaults to None.
            save_with_safetensors (bool, optional): Whether to use safetensors. Defaults to True.
                
        Returns:
            str: Path to saved checkpoint
        """
        if self.fabric is None:
            raise ValueError("Fabric not set up. Call setup_fabric() first.")
        
        # Generate checkpoint filename if not provided
        if filename is None:
            filename = f"model_epoch_{epoch:03d}.pt"
        
        checkpoint_path = os.path.join(self.checkpoint_dir, filename)
        
        # Save checkpoint with Fabric
        state = {
            "model": model.state_dict(),
            "optimizer": optimizer.state_dict(),
            "epoch": epoch,
            "metrics": metrics,
        }
        
        if save_with_safetensors:
            # Save model state dict with safetensors (more secure)
            model_path = os.path.splitext(checkpoint_path)[0] + ".safetensors"
            save_file(model.state_dict(), model_path)
            
            # Save optimizer and metadata separately
            metadata_path = os.path.splitext(checkpoint_path)[0] + "_metadata.json"
            with open(metadata_path, "w") as f:
                json.dump(
                    {
                        "epoch": epoch,
                        "metrics": {k: float(v) if isinstance(v, torch.Tensor) else v for k, v in metrics.items()},
                    }, 
                    f, 
                    indent=2
                )
            
            logger.info(f"Saved checkpoint to {model_path} and {metadata_path}")
            return model_path
        else:
            # Save using Fabric's checkpoint system
            self.fabric.save(checkpoint_path, state)
            logger.info(f"Saved checkpoint to {checkpoint_path}")
            return checkpoint_path
    
    def load_model_checkpoint(
        self, 
        model: torch.nn.Module, 
        optimizer: Optional[torch.optim.Optimizer] = None,
        checkpoint_path: Optional[str] = None,
        load_optimizer: bool = True,
        strict: bool = True,
        from_safetensors: bool = True,
    ) -> Tuple[torch.nn.Module, Optional[torch.optim.Optimizer], Dict[str, Any]]:
        """
        Load model checkpoint with Fabric
        
        Args:
            model (torch.nn.Module): PyTorch model
            optimizer (Optional[torch.optim.Optimizer], optional): Optimizer. Defaults to None.
            checkpoint_path (Optional[str], optional): Path to checkpoint. Defaults to None.
            load_optimizer (bool, optional): Whether to load optimizer. Defaults to True.
            strict (bool, optional): Whether to enforce strict loading. Defaults to True.
            from_safetensors (bool, optional): Whether to load from safetensors. Defaults to True.
                
        Returns:
            Tuple[torch.nn.Module, Optional[torch.optim.Optimizer], Dict[str, Any]]: 
                Loaded model, optimizer, and metadata
        """
        if self.fabric is None:
            raise ValueError("Fabric not set up. Call setup_fabric() first.")
        
        # Find latest checkpoint if path not provided
        if checkpoint_path is None:
            checkpoints = [f for f in os.listdir(self.checkpoint_dir) if f.endswith(".pt") or f.endswith(".safetensors")]
            if not checkpoints:
                raise ValueError(f"No checkpoints found in {self.checkpoint_dir}")
            
            # Sort by modification time
            checkpoints = sorted(
                checkpoints,
                key=lambda x: os.path.getmtime(os.path.join(self.checkpoint_dir, x)),
                reverse=True
            )
            checkpoint_path = os.path.join(self.checkpoint_dir, checkpoints[0])
        
        # Load checkpoint
        if from_safetensors and checkpoint_path.endswith(".safetensors"):
            # Load model state dict from safetensors
            state_dict = load_file(checkpoint_path)
            model.load_state_dict(state_dict, strict=strict)
            
            # Load metadata if available
            metadata = {}
            metadata_path = os.path.splitext(checkpoint_path)[0] + "_metadata.json"
            if os.path.exists(metadata_path):
                with open(metadata_path, "r") as f:
                    metadata = json.load(f)
            
            if load_optimizer and optimizer is not None:
                logger.warning("Optimizer state not available in safetensors format.")
        else:
            # Load using Fabric's checkpoint system
            state = self.fabric.load(checkpoint_path)
            
            # Load model state dict
            model.load_state_dict(state["model"], strict=strict)
            
            # Load optimizer state dict if requested
            if load_optimizer and optimizer is not None and "optimizer" in state:
                optimizer.load_state_dict(state["optimizer"])
            
            # Extract metadata
            metadata = {
                "epoch": state.get("epoch", 0),
                "metrics": state.get("metrics", {})
            }
        
        logger.info(f"Loaded checkpoint from {checkpoint_path}")
        return model, optimizer, metadata


class MemoryOptimization:
    """
    Utilities for memory optimization in deep learning models
    
    Provides methods for analyzing and optimizing memory usage in deep learning models,
    especially useful for large models trained with limited resources.
    """
    
    @staticmethod
    def print_model_memory_usage(model: torch.nn.Module, input_shape: Tuple) -> Dict[str, float]:
        """
        Analyze memory usage of a model
        
        Args:
            model (torch.nn.Module): PyTorch model
            input_shape (Tuple): Input shape for the model
            
        Returns:
            Dict[str, float]: Memory usage statistics in MB
        """
        # Calculate parameter memory
        param_size = 0
        buffer_size = 0
        
        for param in model.parameters():
            param_size += param.nelement() * param.element_size()
        
        for buffer in model.buffers():
            buffer_size += buffer.nelement() * buffer.element_size()
        
        param_size_mb = param_size / (1024 ** 2)
        buffer_size_mb = buffer_size / (1024 ** 2)
        
        # Estimate activation memory
        device = next(model.parameters()).device
        input_tensor = torch.randn(input_shape, device=device)
        
        # Record memory before forward pass
        torch.cuda.reset_peak_memory_stats()
        torch.cuda.empty_cache()
        start_mem = torch.cuda.memory_allocated()
        
        # Forward pass
        with torch.no_grad():
            output = model(input_tensor)
        
        # Record memory after forward pass
        end_mem = torch.cuda.max_memory_allocated()
        activation_memory = end_mem - start_mem
        activation_memory_mb = activation_memory / (1024 ** 2)
        
        # Total memory
        total_memory_mb = param_size_mb + buffer_size_mb + activation_memory_mb
        
        # Print results
        memory_stats = {
            "parameter_memory_mb": param_size_mb,
            "buffer_memory_mb": buffer_size_mb,
            "activation_memory_mb": activation_memory_mb,
            "total_memory_mb": total_memory_mb
        }
        
        logger.info(f"Model Memory Usage:")
        logger.info(f"  Parameters: {param_size_mb:.2f} MB")
        logger.info(f"  Buffers: {buffer_size_mb:.2f} MB")
        logger.info(f"  Activations: {activation_memory_mb:.2f} MB")
        logger.info(f"  Total: {total_memory_mb:.2f} MB")
        
        return memory_stats
    
    @staticmethod
    def apply_activation_checkpointing(model: torch.nn.Module, modules_to_checkpoint: List[type]) -> torch.nn.Module:
        """
        Apply activation checkpointing to reduce memory usage
        
        Args:
            model (torch.nn.Module): PyTorch model
            modules_to_checkpoint (List[type]): List of module types to apply checkpointing to
            
        Returns:
            torch.nn.Module: Model with checkpointing applied
        """
        # Import checkpoint function
        from torch.utils.checkpoint import checkpoint
        
        # Apply checkpointing to specified module types
        for name, module in model.named_children():
            if any(isinstance(module, module_type) for module_type in modules_to_checkpoint):
                # Wrap module's forward method with checkpoint
                original_forward = module.forward
                
                def checkpointed_forward(self, *args, **kwargs):
                    return checkpoint(original_forward, *args, **kwargs)
                
                module.forward = checkpointed_forward.__get__(module)
            
            # Recursively apply to children
            if len(list(module.children())) > 0:
                MemoryOptimization.apply_activation_checkpointing(module, modules_to_checkpoint)
        
        return model
    
    @staticmethod
    def optimize_memory_usage(
        model: torch.nn.Module,
        use_channels_last: bool = True,
        use_compile: bool = True,
        use_activation_checkpointing: bool = False,
        modules_to_checkpoint: Optional[List[type]] = None,
    ) -> torch.nn.Module:
        """
        Apply various memory optimization techniques to a model
        
        Args:
            model (torch.nn.Module): PyTorch model
            use_channels_last (bool, optional): Whether to use channels last memory format. Defaults to True.
            use_compile (bool, optional): Whether to use torch.compile. Defaults to True.
            use_activation_checkpointing (bool, optional): Whether to use activation checkpointing. Defaults to False.
            modules_to_checkpoint (Optional[List[type]], optional): Module types to checkpoint. Defaults to None.
            
        Returns:
            torch.nn.Module: Optimized model
        """
        # Use channels last memory format for convolutional models (improves performance on CUDA)
        if use_channels_last and torch.cuda.is_available():
            model = model.to(memory_format=torch.channels_last)
            logger.info("Using channels last memory format")
        
        # Apply activation checkpointing if requested
        if use_activation_checkpointing and modules_to_checkpoint:
            model = MemoryOptimization.apply_activation_checkpointing(model, modules_to_checkpoint)
            logger.info(f"Applied activation checkpointing to {len(modules_to_checkpoint)} module types")
        
        # Use torch.compile if available and requested (PyTorch 2.0+)
        if use_compile and hasattr(torch, 'compile'):
            model = torch.compile(model)
            logger.info("Applied torch.compile to model")
        
        return model


class ModelShardingHelper:
    """
    Helper for model sharding and checkpoint sharding
    
    This class provides utilities for working with sharded models and checkpoints,
    particularly useful for very large models that don't fit in a single GPU.
    
    Args:
        model_size_gb (float): Estimated model size in GB
        checkpoint_dir (str, optional): Directory for checkpoints. Defaults to "./checkpoints".
    """
    
    def __init__(self, model_size_gb: float, checkpoint_dir: str = "./checkpoints"):
        self.model_size_gb = model_size_gb
        self.checkpoint_dir = checkpoint_dir
        os.makedirs(checkpoint_dir, exist_ok=True)
    
    def create_optimal_fsdp_config(
        self,
        transformer_modules: List[type],
        cpu_offload: bool = False,
    ) -> Dict[str, Any]:
        """
        Create optimal FSDP configuration based on model size and hardware
        
        Args:
            transformer_modules (List[type]): List of transformer module types to wrap
            cpu_offload (bool, optional): Whether to offload to CPU. Defaults to False.
            
        Returns:
            Dict[str, Any]: FSDP configuration
        """
        # Determine sharding strategy based on model size and available GPUs
        num_gpus = torch.cuda.device_count() if torch.cuda.is_available() else 0
        
        # Default to FULL_SHARD for large models
        sharding_strategy = ShardingStrategy.FULL_SHARD
        
        # For smaller models with enough GPUs, use SHARD_GRAD_OP for better performance
        if num_gpus >= 4 and self.model_size_gb < 10:
            sharding_strategy = ShardingStrategy.SHARD_GRAD_OP
        
        # Configure mixed precision
        mp_config = MixedPrecision(
            param_dtype=torch.float16,
            reduce_dtype=torch.float16,
            buffer_dtype=torch.float16,
        )
        
        # Create auto wrap policy
        auto_wrap_policy = transformer_auto_wrap_policy(
            transformer_layer_cls=tuple(transformer_modules)
        )
        
        # Return FSDP configuration
        return {
            "auto_wrap_policy": auto_wrap_policy,
            "cpu_offload": CPUOffload(offload_params=cpu_offload),
            "mixed_precision": mp_config,
            "backward_prefetch": BackwardPrefetch.BACKWARD_PRE,
            "sharding_strategy": sharding_strategy,
            "state_dict_type": StateDictType.FULL_STATE_DICT,
            "limit_all_gathers": True,
        }
    
    def convert_checkpoint_to_sharded(
        self,
        checkpoint_path: str,
        num_shards: int,
        save_with_safetensors: bool = True,
    ) -> List[str]:
        """
        Convert a large checkpoint to sharded checkpoints
        
        Args:
            checkpoint_path (str): Path to the checkpoint
            num_shards (int): Number of shards to create
            save_with_safetensors (bool, optional): Whether to use safetensors. Defaults to True.
            
        Returns:
            List[str]: Paths to sharded checkpoints
        """
        # Load checkpoint
        if checkpoint_path.endswith(".safetensors"):
            state_dict = load_file(checkpoint_path)
        else:
            state_dict = torch.load(checkpoint_path, map_location="cpu")
            if "model" in state_dict:
                state_dict = state_dict["model"]
        
        # Get all keys
        all_keys = list(state_dict.keys())
        keys_per_shard = len(all_keys) // num_shards + (1 if len(all_keys) % num_shards != 0 else 0)
        
        # Create shards
        shard_paths = []
        for i in range(num_shards):
            start_idx = i * keys_per_shard
            end_idx = min((i + 1) * keys_per_shard, len(all_keys))
            shard_keys = all_keys[start_idx:end_idx]
            
            # Create shard state dict
            shard_state_dict = {k: state_dict[k] for k in shard_keys}
            
            # Save shard
            base_path = os.path.splitext(checkpoint_path)[0]
            shard_path = f"{base_path}_shard_{i+1}_of_{num_shards}"
            
            if save_with_safetensors:
                shard_path += ".safetensors"
                save_file(shard_state_dict, shard_path)
            else:
                shard_path += ".pt"
                torch.save(shard_state_dict, shard_path)
            
            shard_paths.append(shard_path)
            logger.info(f"Saved shard {i+1}/{num_shards} with {len(shard_keys)} keys to {shard_path}")
        
        return shard_paths
    
    def load_sharded_checkpoint(
        self,
        model: torch.nn.Module,
        shard_paths: List[str],
        strict: bool = False,
    ) -> torch.nn.Module:
        """
        Load a model from sharded checkpoints
        
        Args:
            model (torch.nn.Module): PyTorch model
            shard_paths (List[str]): Paths to sharded checkpoints
            strict (bool, optional): Whether to strictly enforce loading. Defaults to False.
            
        Returns:
            torch.nn.Module: Model with loaded weights
        """
        # Create empty state dict
        state_dict = {}
        
        # Load each shard
        for shard_path in shard_paths:
            logger.info(f"Loading shard from {shard_path}")
            
            if shard_path.endswith(".safetensors"):
                shard_state_dict = load_file(shard_path)
            else:
                shard_state_dict = torch.load(shard_path, map_location="cpu")
                if "model" in shard_state_dict:
                    shard_state_dict = shard_state_dict["model"]
            
            # Update combined state dict
            state_dict.update(shard_state_dict)
        
        # Load state dict into model
        model.load_state_dict(state_dict, strict=strict)
        logger.info(f"Successfully loaded model from {len(shard_paths)} shards with {len(state_dict)} parameters")
        
        return model


class BatchSizeOptimizer:
    """
    Utility to find the optimal batch size for a model
    
    This class automatically finds the optimal batch size for a model
    based on available GPU memory and model characteristics.
    
    Args:
        model (torch.nn.Module): PyTorch model
        input_shape_template (Tuple): Template for input shape (without batch dimension)
        device (str, optional): Device to use. Defaults to "cuda".
        start_batch_size (int, optional): Starting batch size. Defaults to 1.
        max_batch_size (int, optional): Maximum batch size to try. Defaults to 512.
        max_memory_usage (float, optional): Maximum fraction of GPU memory to use. Defaults to 0.85.
    """
    
    def __init__(
        self,
        model: torch.nn.Module,
        input_shape_template: Tuple,
        device: str = "cuda",
        start_batch_size: int = 1,
        max_batch_size: int = 512,
        max_memory_usage: float = 0.85,
    ):
        self.model = model
        self.input_shape_template = input_shape_template
        self.device = device
        self.start_batch_size = start_batch_size
        self.max_batch_size = max_batch_size
        self.max_memory_usage = max_memory_usage
    
    def find_optimal_batch_size(
        self, 
        optimization_metric: str = "memory",
        verbose: bool = True,
    ) -> int:
        """
        Find the optimal batch size
        
        Args:
            optimization_metric (str, optional): What to optimize for - "memory" or "speed". 
                Defaults to "memory".
            verbose (bool, optional): Whether to print progress. Defaults to True.
            
        Returns:
            int: Optimal batch size
        """
        if not torch.cuda.is_available() and self.device == "cuda":
            logger.warning("CUDA not available, using CPU. This will be slow.")
            self.device = "cpu"
        
        # Move model to device
        self.model = self.model.to(self.device)
        self.model.eval()
        
        # Binary search for maximum batch size
        low, high = self.start_batch_size, self.max_batch_size
        optimal_batch_size = low
        
        # For speed optimization, we'll also track timing
        best_throughput = 0
        
        while low <= high:
            mid = (low + high) // 2
            
            if verbose:
                logger.info(f"Trying batch size: {mid}")
            
            try:
                # Reset memory stats
                if self.device == "cuda":
                    torch.cuda.reset_peak_memory_stats()
                    torch.cuda.empty_cache()
                
                # Create input tensor
                input_tensor = torch.randn((mid,) + self.input_shape_template, device=self.device)
                
                # Time the forward pass
                start_time = time.time()
                
                # Forward pass
                with torch.no_grad():
                    _ = self.model(input_tensor)
                
                # Calculate throughput
                elapsed = time.time() - start_time
                throughput = mid / elapsed
                
                if verbose:
                    logger.info(f"  Success - Throughput: {throughput:.2f} samples/sec")
                
                # Check memory usage
                if self.device == "cuda":
                    peak_memory = torch.cuda.max_memory_allocated()
                    total_memory = torch.cuda.get_device_properties(0).total_memory
                    memory_usage = peak_memory / total_memory
                    
                    if verbose:
                        logger.info(f"  Memory usage: {memory_usage:.2%}")
                    
                    # If we're over our memory threshold, we should go lower
                    if memory_usage > self.max_memory_usage:
                        if verbose:
                            logger.info("  Memory usage too high, reducing batch size")
                        high = mid - 1
                        continue
                
                # If we're optimizing for speed
                if optimization_metric == "speed" and throughput > best_throughput:
                    best_throughput = throughput
                    optimal_batch_size = mid
                else:
                    # For memory optimization, always prefer larger batch size
                    optimal_batch_size = mid
                
                # Try a larger batch size
                low = mid + 1
                
            except RuntimeError as e:
                # Check if it's an out-of-memory error
                if "CUDA out of memory" in str(e) or "DefaultCPUAllocator: not enough memory" in str(e):
                    if verbose:
                        logger.info(f"  Out of memory with batch size {mid}, reducing")
                    # Try a smaller batch size
                    high = mid - 1
                else:
                    # Some other error
                    raise e
        
        # Suggest a slightly smaller batch size for stability
        if optimization_metric == "memory" and optimal_batch_size > 1:
            # Reduce by 5% to provide some headroom
            optimal_batch_size = max(1, int(optimal_batch_size * 0.95))
        
        logger.info(f"Optimal batch size: {optimal_batch_size}")
        return optimal_batch_size


class SecurityWrapper:
    """
    Security utilities for model serialization and deserialization
    
    This class provides secure methods for saving and loading models
    to prevent vulnerabilities related to insecure deserialization.
    
    Args:
        base_dir (str, optional): Base directory for saving models. Defaults to "./secure_models".
    """
    
    def __init__(self, base_dir: str = "./secure_models"):
        self.base_dir = base_dir
        os.makedirs(base_dir, exist_ok=True)
    
    def save_model_secure(
        self,
        model: torch.nn.Module,
        metadata: Dict[str, Any],
        filename: str,
    ) -> str:
        """
        Save model securely using safetensors
        
        Args:
            model (torch.nn.Module): PyTorch model
            metadata (Dict[str, Any]): Metadata to save
            filename (str): Filename for the model
            
        Returns:
            str: Path to saved model
        """
        # Ensure we're using safetensors
        model_path = os.path.join(self.base_dir, filename)
        if not model_path.endswith(".safetensors"):
            model_path += ".safetensors"
        
        # Get state dict
        state_dict = model.state_dict()
        
        # Save model state dict
        save_file(state_dict, model_path)
        
        # Save metadata separately as JSON
        metadata_path = os.path.splitext(model_path)[0] + "_metadata.json"
        with open(metadata_path, "w") as f:
            json.dump(
                {k: (float(v) if isinstance(v, torch.Tensor) else v) for k, v in metadata.items()}, 
                f, 
                indent=2
            )
        
        logger.info(f"Securely saved model to {model_path} and metadata to {metadata_path}")
        return model_path
    
    def load_model_secure(
        self,
        model: torch.nn.Module,
        filename: str,
        strict: bool = True,
    ) -> Tuple[torch.nn.Module, Dict[str, Any]]:
        """
        Load model securely using safetensors
        
        Args:
            model (torch.nn.Module): PyTorch model
            filename (str): Filename of the model
            strict (bool, optional): Whether to strictly enforce loading. Defaults to True.
            
        Returns:
            Tuple[torch.nn.Module, Dict[str, Any]]: Loaded model and metadata
        """
        # Ensure we're using safetensors
        model_path = os.path.join(self.base_dir, filename)
        if not model_path.endswith(".safetensors"):
            model_path += ".safetensors"
        
        # Check if file exists
        if not os.path.exists(model_path):
            raise FileNotFoundError(f"Model file not found: {model_path}")
        
        # Load model state dict
        state_dict = load_file(model_path)
        
        # Load model
        model.load_state_dict(state_dict, strict=strict)
        
        # Load metadata if available
        metadata = {}
        metadata_path = os.path.splitext(model_path)[0] + "_metadata.json"
        if os.path.exists(metadata_path):
            with open(metadata_path, "r") as f:
                metadata = json.load(f)
        
        logger.info(f"Securely loaded model from {model_path}")
        return model, metadata



"""
Performance profiling and benchmarking utilities
"""
import os
import time
import json
import logging
import platform
from typing import Dict, List, Any, Optional, Union, Tuple, Callable

import torch
import numpy as np
import matplotlib.pyplot as plt
from torch.utils.data import DataLoader
import lightning as L

logger = logging.getLogger(__name__)


class ModelProfiler:
    """
    Utility for profiling PyTorch models
    
    This class provides methods for profiling model performance, 
    including inference time, memory usage, and parameter counts.
    
    Args:
        model (torch.nn.Module): PyTorch model to profile
        input_shape (Tuple): Input shape for the model (including batch dimension)
        device (str, optional): Device to run profiling on. Defaults to "cuda" if available.
        log_dir (str, optional): Directory to save profile results. Defaults to "./profiles".
    """
    
    def __init__(
        self, 
        model: torch.nn.Module, 
        input_shape: Tuple,
        device: Optional[str] = None,
        log_dir: str = "./profiles"
    ):
        self.model = model
        self.input_shape = input_shape
        self.device = device or ("cuda" if torch.cuda.is_available() else "cpu")
        self.log_dir = log_dir
        
        os.makedirs(log_dir, exist_ok=True)
    
    def profile_inference_time(
        self, 
        num_warmup: int = 10, 
        num_runs: int = 100,
        batch_size: Optional[int] = None,
    ) -> Dict[str, float]:
        """
        Profile model inference time
        
        Args:
            num_warmup (int, optional): Number of warmup runs. Defaults to 10.
            num_runs (int, optional): Number of timed runs. Defaults to 100.
            batch_size (Optional[int], optional): Custom batch size. Defaults to None
                (uses batch size from input_shape).
                
        Returns:
            Dict[str, float]: Inference time statistics in milliseconds
        """
        # Move model to device and set to eval mode
        model = self.model.to(self.device)
        model.eval()
        
        # Create input tensor
        if batch_size is not None:
            input_shape = (batch_size,) + self.input_shape[1:]
        else:
            input_shape = self.input_shape
        
        input_tensor = torch.randn(input_shape, device=self.device)
        
        # Warmup runs
        with torch.no_grad():
            for _ in range(num_warmup):
                _ = model(input_tensor)
        
        # Synchronize before timing
        if self.device == "cuda":
            torch.cuda.synchronize()
        
        # Timed runs
        times = []
        with torch.no_grad():
            for _ in range(num_runs):
                start_time = time.time()
                _ = model(input_tensor)
                
                if self.device == "cuda":
                    torch.cuda.synchronize()
                
                end_time = time.time()
                times.append((end_time - start_time) * 1000)  # Convert to ms
        
        # Calculate statistics
        times = np.array(times)
        stats = {
            "mean_ms": float(np.mean(times)),
            "median_ms": float(np.median(times)),
            "min_ms": float(np.min(times)),
            "max_ms": float(np.max(times)),
            "std_ms": float(np.std(times)),
            "p95_ms": float(np.percentile(times, 95)),
            "p99_ms": float(np.percentile(times, 99)),
            "samples_per_second": float(input_shape[0] * 1000 / np.mean(times)),
            "batch_size": input_shape[0],
        }
        
        # Print summary
        logger.info(f"Inference Time Summary (batch size={input_shape[0]}):")
        logger.info(f"  Mean: {stats['mean_ms']:.2f} ms")
        logger.info(f"  Median: {stats['median_ms']:.2f} ms")
        logger.info(f"  Min: {stats['min_ms']:.2f} ms")
        logger.info(f"  Max: {stats['max_ms']:.2f} ms")
        logger.info(f"  P95: {stats['p95_ms']:.2f} ms")
        logger.info(f"  P99: {stats['p99_ms']:.2f} ms")
        logger.info(f"  Throughput: {stats['samples_per_second']:.2f} samples/sec")
        
        return stats
    
    def profile_memory_usage(self) -> Dict[str, float]:
        """
        Profile model memory usage
        
        Returns:
            Dict[str, float]: Memory usage statistics in MB
        """
        # Move model to device and set to eval mode
        model = self.model.to(self.device)
        model.eval()
        
        # Create input tensor
        input_tensor = torch.randn(self.input_shape, device=self.device)
        
        # Calculate parameter memory
        param_size = 0
        for param in model.parameters():
            param_size += param.nelement() * param.element_size()
        
        param_size_mb = param_size / (1024 ** 2)
        
        # Empty cache and record memory before forward pass
        if self.device == "cuda":
            torch.cuda.empty_cache()
            torch.cuda.reset_peak_memory_stats()
            start_mem = torch.cuda.memory_allocated()
        
        # Forward pass
        with torch.no_grad():
            _ = model(input_tensor)
        
        # Record memory after forward pass
        if self.device == "cuda":
            end_mem = torch.cuda.max_memory_allocated()
            activation_memory = end_mem - start_mem
            activation_memory_mb = activation_memory / (1024 ** 2)
            
            # Get total GPU memory
            total_gpu_memory = torch.cuda.get_device_properties(0).total_memory
            total_gpu_memory_mb = total_gpu_memory / (1024 ** 2)
        else:
            # CPU memory estimation is less accurate
            activation_memory_mb = 0  # Not easily measurable on CPU
            total_gpu_memory_mb = 0
        
        # Count number of parameters
        total_params = sum(p.numel() for p in model.parameters())
        trainable_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
        
        # Calculate total memory
        total_memory_mb = param_size_mb + activation_memory_mb
        
        # Create memory stats
        memory_stats = {
            "parameter_memory_mb": param_size_mb,
            "activation_memory_mb": activation_memory_mb,
            "total_memory_mb": total_memory_mb,
            "total_parameters": total_params,
            "trainable_parameters": trainable_params,
            "total_gpu_memory_mb": total_gpu_memory_mb,
            "memory_utilization": total_memory_mb / total_gpu_memory_mb if total_gpu_memory_mb > 0 else 0,
        }
        
        # Print summary
        logger.info(f"Memory Usage Summary:")
        logger.info(f"  Parameters: {total_params:,} ({memory_stats['parameter_memory_mb']:.2f} MB)")
        logger.info(f"  Activations: {memory_stats['activation_memory_mb']:.2f} MB")
        logger.info(f"  Total: {memory_stats['total_memory_mb']:.2f} MB")
        if self.device == "cuda":
            logger.info(f"  GPU Memory Utilization: {memory_stats['memory_utilization']:.2%}")
        
        return memory_stats
    
    def profile_layer_compute_time(
        self, 
        num_runs: int = 10, 
        return_per_layer: bool = False,
    ) -> Dict[str, Union[float, Dict[str, float]]]:
        """
        Profile computation time for each layer
        
        Args:
            num_runs (int, optional): Number of runs. Defaults to 10.
            return_per_layer (bool, optional): Whether to return per-layer times. Defaults to False.
            
        Returns:
            Dict[str, Union[float, Dict[str, float]]]: Layer timing statistics
        """
        # Move model to device and set to eval mode
        model = self.model.to(self.device)
        model.eval()
        
        # Create input tensor
        input_tensor = torch.randn(self.input_shape, device=self.device)
        
        # Dictionary to store times
        layer_times = {}
        
        # Define forward hooks
        def hook_fn(name):
            def hook(module, input, output):
                if self.device == "cuda":
                    torch.cuda.synchronize()
                end_time = time.time()
                if name in layer_times:
                    layer_times[name].append(end_time)
                else:
                    layer_times[name] = [end_time]
                return None
            return hook
        
        def pre_hook_fn(name):
            def hook(module, input):
                if self.device == "cuda":
                    torch.cuda.synchronize()
                start_time = time.time()
                if name + "_pre" in layer_times:
                    layer_times[name + "_pre"].append(start_time)
                else:
                    layer_times[name + "_pre"] = [start_time]
                return None
            return hook
        
        # Register hooks
        hooks = []
        pre_hooks = []
        for name, module in model.named_modules():
            if name:  # Skip the root module
                hooks.append(module.register_forward_hook(hook_fn(name)))
                pre_hooks.append(module.register_forward_pre_hook(pre_hook_fn(name)))
        
        # Multiple forward passes
        with torch.no_grad():
            for _ in range(num_runs):
                _ = model(input_tensor)
        
        # Remove hooks
        for hook in hooks:
            hook.remove()
        for hook in pre_hooks:
            hook.remove()
        
        # Calculate time per layer
        layer_compute_times = {}
        for name in layer_times:
            if name.endswith("_pre"):
                base_name = name[:-4]
                if base_name in layer_times:
                    times = []
                    pre_times = layer_times[name]
                    post_times = layer_times[base_name]
                    for i in range(1, len(pre_times)):  # Skip first run (warmup)
                        times.append((post_times[i] - pre_times[i]) * 1000)  # ms
                    layer_compute_times[base_name] = times
        
        # Calculate statistics
        total_time = 0
        layer_stats = {}
        for name, times in layer_compute_times.items():
            mean_time = np.mean(times) if times else 0
            total_time += mean_time
            layer_stats[name] = {
                "mean_ms": float(mean_time),
                "percent": 0,  # Will fill later
                "max_ms": float(np.max(times)) if times else 0,
                "min_ms": float(np.min(times)) if times else 0,
            }
        
        # Calculate percentages
        for name in layer_stats:
            layer_stats[name]["percent"] = layer_stats[name]["mean_ms"] / total_time if total_time > 0 else 0
        
        # Sort by compute time
        sorted_layers = sorted(
            layer_stats.items(), 
            key=lambda x: x[1]["mean_ms"], 
            reverse=True
        )
        
        # Print top 10 most expensive layers
        logger.info(f"Top 10 most compute-intensive layers:")
        for i, (name, stats) in enumerate(sorted_layers[:10]):
            logger.info(f"  {i+1}. {name}: {stats['mean_ms']:.2f} ms ({stats['percent']:.2%})")
        
        result = {
            "total_compute_time_ms": float(total_time),
        }
        
        if return_per_layer:
            result["layer_stats"] = layer_stats
        
        return result
    
    def generate_profile_report(
        self, 
        include_layer_profile: bool = True,
        batch_sizes: Optional[List[int]] = None,
    ) -> str:
        """
        Generate comprehensive profile report
        
        Args:
            include_layer_profile (bool, optional): Whether to include layer-wise profiling. 
                Defaults to True.
            batch_sizes (Optional[List[int]], optional): Batch sizes to profile. 
                Defaults to None (uses [1, 2, 4, 8, 16, 32] if possible).
                
        Returns:
            str: Path to saved report
        """
        # Get hardware info
        system_info = {
            "platform": platform.platform(),
            "python_version": platform.python_version(),
            "torch_version": torch.__version__,
            "cuda_version": torch.version.cuda if torch.cuda.is_available() else "N/A",
            "device": self.device,
        }
        
        if self.device == "cuda":
            system_info["gpu"] = torch.cuda.get_device_name(0)
            system_info["gpu_count"] = torch.cuda.device_count()
        
        # Profile memory
        memory_stats = self.profile_memory_usage()
        
        # Profile batch sizes
        if batch_sizes is None:
            original_batch_size = self.input_shape[0]
            batch_sizes = []
            for bs in [1, 2, 4, 8, 16, 32]:
                if bs <= original_batch_size * 2:  # Don't go too far beyond original batch size
                    batch_sizes.append(bs)
        
        batch_size_results = {}
        for bs in batch_sizes:
            try:
                batch_size_results[bs] = self.profile_inference_time(
                    num_warmup=5, 
                    num_runs=20, 
                    batch_size=bs
                )
            except RuntimeError as e:
                logger.warning(f"Failed to profile batch size {bs}: {e}")
        
        # Profile layer compute time if requested
        layer_stats = None
        if include_layer_profile:
            result = self.profile_layer_compute_time(return_per_layer=True)
            layer_stats = result.get("layer_stats", None)
        
        # Compile results
        profile_results = {
            "timestamp": time.strftime("%Y-%m-%d %H:%M:%S"),
            "system_info": system_info,
            "model_name": self.model.__class__.__name__,
            "input_shape": list(self.input_shape),
            "memory_stats": memory_stats,
            "batch_size_results": batch_size_results,
            "layer_stats": layer_stats,
        }
        
        # Save results
        report_path = os.path.join(
            self.log_dir, 
            f"profile_{self.model.__class__.__name__}_{time.strftime('%Y%m%d_%H%M%S')}.json"
        )
        
        with open(report_path, "w") as f:
            json.dump(profile_results, f, indent=2)
        
        logger.info(f"Profile report saved to {report_path}")
        
        # Generate plots
        self._generate_profile_plots(profile_results)
        
        return report_path
    
    def _generate_profile_plots(self, profile_results: Dict[str, Any]) -> None:
        """
        Generate plots from profile results
        
        Args:
            profile_results (Dict[str, Any]): Profile results
        """
        # Get model name for plot titles
        model_name = profile_results["model_name"]
        
        # Create batch size vs throughput plot
        if profile_results["batch_size_results"]:
            plt.figure(figsize=(10, 6))
            
            batch_sizes = sorted([int(bs) for bs in profile_results["batch_size_results"].keys()])
            throughput = [profile_results["batch_size_results"][str(bs)]["samples_per_second"] for bs in batch_sizes]
            latency = [profile_results["batch_size_results"][str(bs)]["mean_ms"] for bs in batch_sizes]
            
            # Plot throughput
            plt.subplot(1, 2, 1)
            plt.plot(batch_sizes, throughput, 'o-', color='blue')
            plt.xlabel('Batch Size')
            plt.ylabel('Throughput (samples/sec)')
            plt.title(f'{model_name} Throughput vs Batch Size')
            plt.grid(True)
            
            # Plot latency
            plt.subplot(1, 2, 2)
            plt.plot(batch_sizes, latency, 'o-', color='red')
            plt.xlabel('Batch Size')
            plt.ylabel('Latency (ms)')
            plt.title(f'{model_name} Latency vs Batch Size')
            plt.grid(True)
            
            plt.tight_layout()
            plt.savefig(os.path.join(self.log_dir, f"{model_name}_batch_performance.png"))
            plt.close()
        
        # Create layer compute time plot
        if profile_results["layer_stats"]:
            # Sort layers by compute time
            layer_stats = profile_results["layer_stats"]
            sorted_layers = sorted(
                layer_stats.items(), 
                key=lambda x: x[1]["mean_ms"], 
                reverse=True
            )
            
            # Plot top 15 layers
            top_n = min(15, len(sorted_layers))
            top_layers = sorted_layers[:top_n]
            
            plt.figure(figsize=(12, 6))
            
            names = [layer[0].split(".")[-1] for layer in top_layers]  # Use only last part of name
            compute_times = [layer[1]["mean_ms"] for layer in top_layers]
            percentages = [layer[1]["percent"] for layer in top_layers]
            
            # Create horizontal bar chart
            y_pos = range(len(names))
            plt.barh(y_pos, compute_times, align='center', color='skyblue')
            
            # Add percentage labels
            for i, (time, pct) in enumerate(zip(compute_times, percentages)):
                plt.text(time + 0.1, i, f"{time:.2f} ms ({pct:.1%})", va='center')
            
            plt.yticks(y_pos, names)
            plt.xlabel('Compute Time (ms)')
            plt.title(f'Top {top_n} Most Compute-Intensive Layers in {model_name}')
            plt.tight_layout()
            
            plt.savefig(os.path.join(self.log_dir, f"{model_name}_layer_profile.png"))
            plt.close()


class ArchitectureBenchmarker:
    """
    Benchmark different neural network architectures
    
    This class allows easy benchmarking of multiple model architectures
    to compare performance metrics.
    
    Args:
        models (Dict[str, torch.nn.Module]): Dictionary of models to benchmark
        input_shape (Tuple): Input shape for models (including batch dimension)
        device (str, optional): Device to run benchmarks on. Defaults to "cuda" if available.
        log_dir (str, optional): Directory to save benchmark results. Defaults to "./benchmarks".
    """
    
    def __init__(
        self, 
        models: Dict[str, torch.nn.Module],
        input_shape: Tuple,
        device: Optional[str] = None,
        log_dir: str = "./benchmarks"
    ):
        self.models = models
        self.input_shape = input_shape
        self.device = device or ("cuda" if torch.cuda.is_available() else "cpu")
        self.log_dir = log_dir
        
        os.makedirs(log_dir, exist_ok=True)
    
    def run_benchmarks(
        self,
        benchmark_memory: bool = True,
        benchmark_inference: bool = True,
        benchmark_throughput: bool = True,
        batch_size: Optional[int] = None,
    ) -> Dict[str, Dict[str, Any]]:
        """
        Run comprehensive benchmarks on all models
        
        Args:
            benchmark_memory (bool, optional): Whether to benchmark memory usage. Defaults to True.
            benchmark_inference (bool, optional): Whether to benchmark inference time. Defaults to True.
            benchmark_throughput (bool, optional): Whether to benchmark throughput. Defaults to True.
            batch_size (Optional[int], optional): Custom batch size for throughput test. 
                Defaults to None (uses batch size from input_shape).
                
        Returns:
            Dict[str, Dict[str, Any]]: Benchmark results per model
        """
        results = {}
        
        for name, model in self.models.items():
            logger.info(f"Benchmarking {name}...")
            model_results = {}
            profiler = ModelProfiler(model, self.input_shape, self.device, self.log_dir)
            
            # Benchmark memory
            if benchmark_memory:
                model_results["memory"] = profiler.profile_memory_usage()
            
            # Benchmark inference
            if benchmark_inference:
                model_results["inference"] = profiler.profile_inference_time()
            
            # Benchmark throughput with different batch size
            if benchmark_throughput and batch_size is not None:
                model_results["throughput"] = profiler.profile_inference_time(batch_size=batch_size)
            
            results[name] = model_results
        
        # Save results
        benchmark_path = os.path.join(
            self.log_dir, 
            f"benchmark_results_{time.strftime('%Y%m%d_%H%M%S')}.json"
        )
        
        with open(benchmark_path, "w") as f:
            # Convert any non-serializable values to strings or numbers
            serializable_results = json.loads(
                json.dumps(results, default=lambda o: float(o) if isinstance(o, torch.Tensor) else str(o))
            )
            json.dump(serializable_results, f, indent=2)
        
        logger.info(f"Benchmark results saved to {benchmark_path}")
        
        # Generate comparison plots
        self._generate_comparison_plots(results)
        
        return results
    
    def _generate_comparison_plots(self, results: Dict[str, Dict[str, Any]]) -> None:
        """
        Generate comparison plots from benchmark results
        
        Args:
            results (Dict[str, Dict[str, Any]]): Benchmark results
        """
        # Plot memory comparison
        if all("memory" in model_results for model_results in results.values()):
            plt.figure(figsize=(12, 6))
            
            model_names = list(results.keys())
            param_memory = [results[name]["memory"]["parameter_memory_mb"] for name in model_names]
            activation_memory = [results[name]["memory"]["activation_memory_mb"] for name in model_names]
            
            x = range(len(model_names))
            width = 0.35
            
            plt.bar(x, param_memory, width, label='Parameter Memory', color='skyblue')
            plt.bar([i + width for i in x], activation_memory, width, label='Activation Memory', color='orange')
            
            plt.xlabel('Model')
            plt.ylabel('Memory (MB)')
            plt.title('Memory Usage Comparison')
            plt.xticks([i + width/2 for i in x], model_names, rotation=45, ha='right')
            plt.legend()
            plt.tight_layout()
            
            plt.savefig(os.path.join(self.log_dir, "memory_comparison.png"))
            plt.close()
        
        # Plot inference time comparison
        if all("inference" in model_results for model_results in results.values()):
            plt.figure(figsize=(12, 6))
            
            model_names = list(results.keys())
            inference_times = [results[name]["inference"]["mean_ms"] for name in model_names]
            
            plt.bar(model_names, inference_times, color='skyblue')
            plt.xlabel('Model')
            plt.ylabel('Inference Time (ms)')
            plt.title('Inference Time Comparison')
            plt.xticks(rotation=45, ha='right')
            
            # Add labels
            for i, v in enumerate(inference_times):
                plt.text(i, v + 0.5, f"{v:.2f} ms", ha='center')
            
            plt.tight_layout()
            
            plt.savefig(os.path.join(self.log_dir, "inference_comparison.png"))
            plt.close()
        
        # Plot throughput comparison
        if all("throughput" in model_results for model_results in results.values()):
            plt.figure(figsize=(12, 6))
            
            model_names = list(results.keys())
            throughput = [results[name]["throughput"]["samples_per_second"] for name in model_names]
            
            plt.bar(model_names, throughput, color='skyblue')
            plt.xlabel('Model')
            plt.ylabel('Throughput (samples/sec)')
            plt.title('Throughput Comparison')
            plt.xticks(rotation=45, ha='right')
            
            # Add labels
            for i, v in enumerate(throughput):
                plt.text(i, v + 0.5, f"{v:.2f}", ha='center')
            
            plt.tight_layout()
            
            plt.savefig(os.path.join(self.log_dir, "throughput_comparison.png"))
            plt.close()


class LightningProfileCallback(L.Callback):
    """
    Lightning callback for model profiling
    
    This callback profiles model performance during training,
    including iteration time, memory usage, and throughput.
    
    Args:
        profile_every_n_steps (int, optional): How often to profile. Defaults to 100.
        log_dir (str, optional): Directory to save profiles. Defaults to "./profiles".
        profile_memory (bool, optional): Whether to profile memory. Defaults to True.
        profile_batch_scaling (bool, optional): Whether to profile batch scaling. Defaults to False.
    """
    
    def __init__(
        self,
        profile_every_n_steps: int = 100,
        log_dir: str = "./profiles",
        profile_memory: bool = True,
        profile_batch_scaling: bool = False,
    ):
        self.profile_every_n_steps = profile_every_n_steps
        self.log_dir = log_dir
        self.profile_memory = profile_memory
        self.profile_batch_scaling = profile_batch_scaling
        
        os.makedirs(log_dir, exist_ok=True)
        
        self.step_times = []
        self.memory_usage = []
        self.throughput = []
        self.last_time = None
    
    def on_train_start(self, trainer: L.Trainer, pl_module: L.LightningModule) -> None:
        """Log initial model state"""
        if self.profile_memory:
            self._log_memory_usage(trainer, pl_module)
    
    def on_train_batch_start(self, trainer: L.Trainer, pl_module: L.LightningModule, *args, **kwargs) -> None:
        """Record start time of batch"""
        self.last_time = time.time()
    
    def on_train_batch_end(self, trainer: L.Trainer, pl_module: L.LightningModule, *args, **kwargs) -> None:
        """Record end time of batch and calculate metrics"""
        if self.last_time is None:
            return
        
        # Calculate step time
        step_time = time.time() - self.last_time
        self.step_times.append(step_time)
        
        # Calculate throughput
        batch_size = trainer.train_dataloader.batch_size
        throughput = batch_size / step_time
        self.throughput.append(throughput)
        
        # Log memory usage
        if self.profile_memory:
            self._log_memory_usage(trainer, pl_module)
        
        # Profile at specified intervals
        if trainer.global_step % self.profile_every_n_steps == 0:
            self._log_profile(trainer, pl_module)
    
    def on_train_end(self, trainer: L.Trainer, pl_module: L.LightningModule) -> None:
        """Log final profile summary"""
        self._save_profiles(trainer, pl_module)
    
    def _log_memory_usage(self, trainer: L.Trainer, pl_module: L.LightningModule) -> None:
        """Log current memory usage"""
        if torch.cuda.is_available():
            allocated = torch.cuda.memory_allocated() / (1024 ** 2)
            reserved = torch.cuda.memory_reserved() / (1024 ** 2)
            max_allocated = torch.cuda.max_memory_allocated() / (1024 ** 2)
            
            memory_info = {
                "allocated_mb": allocated,
                "reserved_mb": reserved,
                "max_allocated_mb": max_allocated,
                "step": trainer.global_step,
            }
            
            self.memory_usage.append(memory_info)
    
    def _log_profile(self, trainer: L.Trainer, pl_module: L.LightningModule) -> None:
        """Log current profile data"""
        # Calculate statistics
        recent_times = self.step_times[-min(100, len(self.step_times)):]
        recent_throughput = self.throughput[-min(100, len(self.throughput)):]
        
        profile_data = {
            "step": trainer.global_step,
            "iteration_time": {
                "mean_ms": float(np.mean(recent_times) * 1000),
                "median_ms": float(np.median(recent_times) * 1000),
                "min_ms": float(np.min(recent_times) * 1000),
                "max_ms": float(np.max(recent_times) * 1000),
                "p95_ms": float(np.percentile(recent_times, 95) * 1000),
            },
            "throughput": {
                "mean": float(np.mean(recent_throughput)),
                "median": float(np.median(recent_throughput)),
                "min": float(np.min(recent_throughput)),
                "max": float(np.max(recent_throughput)),
            },
        }
        
        # Log to Lightning's logger
        if trainer.logger:
            trainer.logger.log_metrics({
                "iteration_time": profile_data["iteration_time"]["mean_ms"],
                "throughput": profile_data["throughput"]["mean"],
            }, step=trainer.global_step)
        
        # Profile batch scaling if enabled
        if self.profile_batch_scaling and trainer.global_step > 0 and trainer.global_step % (self.profile_every_n_steps * 5) == 0:
            try:
                device = pl_module.device
                batch_profiler = ModelProfiler(pl_module, (1,) + trainer.train_dataloader.dataset[0][0].shape, device, self.log_dir)
                batch_sizes = [1, 2, 4, 8, 16]
                
                for bs in batch_sizes:
                    try:
                        profile_data[f"batch_{bs}"] = batch_profiler.profile_inference_time(
                            num_warmup=3, 
                            num_runs=10, 
                            batch_size=bs
                        )
                    except RuntimeError:
                        break
            except Exception as e:
                logger.warning(f"Failed to profile batch scaling: {e}")
        
        logger.info(f"Profile at step {trainer.global_step}:")
        logger.info(f"  Iteration time: {profile_data['iteration_time']['mean_ms']:.2f} ms")
        logger.info(f"  Throughput: {profile_data['throughput']['mean']:.2f} samples/sec")
    
    def _save_profiles(self, trainer: L.Trainer, pl_module: L.LightningModule) -> None:
        """Save profile data to disk"""
        # Calculate overall statistics
        if self.step_times:
            iteration_time = {
                "mean_ms": float(np.mean(self.step_times) * 1000),
                "median_ms": float(np.median(self.step_times) * 1000),
                "min_ms": float(np.min(self.step_times) * 1000),
                "max_ms": float(np.max(self.step_times) * 1000),
                "p95_ms": float(np.percentile(self.step_times, 95) * 1000),
            }
        else:
            iteration_time = {}
        
        if self.throughput:
            throughput = {
                "mean": float(np.mean(self.throughput)),
                "median": float(np.median(self.throughput)),
                "min": float(np.min(self.throughput)),
                "max": float(np.max(self.throughput)),
                "p95": float(np.percentile(self.throughput, 95)),
            }
        else:
            throughput = {}
        
        # Compile profile data
        profile_data = {
            "model_name": pl_module.__class__.__name__,
            "final_step": trainer.global_step,
            "total_iterations": len(self.step_times),
            "iteration_time": iteration_time,
            "throughput": throughput,
            "memory_usage": self.memory_usage if self.profile_memory else None,
        }
        
        # Save to disk
        profile_path = os.path.join(
            self.log_dir, 
            f"training_profile_{pl_module.__class__.__name__}_{time.strftime('%Y%m%d_%H%M%S')}.json"
        )
        
        with open(profile_path, "w") as f:
            json.dump(profile_data, f, indent=2)
        
        logger.info(f"Training profile saved to {profile_path}")
        
        # Generate plots
        self._generate_profile_plots(profile_data)
    
    def _generate_profile_plots(self, profile_data: Dict[str, Any]) -> None:
        """Generate plots from profile data"""
        model_name = profile_data["model_name"]
        
        # Plot iteration time and throughput over time
        if profile_data.get("memory_usage"):
            plt.figure(figsize=(12, 8))
            
            # Extract data
            steps = [entry["step"] for entry in profile_data["memory_usage"]]
            allocated = [entry["allocated_mb"] for entry in profile_data["memory_usage"]]
            reserved = [entry["reserved_mb"] for entry in profile_data["memory_usage"]]
            max_allocated = [entry["max_allocated_mb"] for entry in profile_data["memory_usage"]]
            
            # Plot memory
            plt.subplot(2, 1, 1)
            plt.plot(steps, allocated, label='Allocated')
            plt.plot(steps, reserved, label='Reserved')
            plt.plot(steps, max_allocated, label='Max Allocated')
            plt.xlabel('Training Step')
            plt.ylabel('Memory (MB)')
            plt.title(f'{model_name} Memory Usage During Training')
            plt.legend()
            plt.grid(True)
            
            # Plot iteration time
            if len(self.step_times) >= len(steps):
                iter_times = self.step_times[:len(steps)]
                throughputs = self.throughput[:len(steps)]
                
                # Create moving average
                window = min(20, len(iter_times))
                iter_times_smooth = np.convolve(
                    iter_times, 
                    np.ones(window) / window, 
                    mode='valid'
                ) * 1000  # Convert to ms
                
                throughput_smooth = np.convolve(
                    throughputs, 
                    np.ones(window) / window, 
                    mode='valid'
                )
                
                # Plot smoothed iteration time
                plt.subplot(2, 1, 2)
                plt.plot(steps[window-1:], iter_times_smooth, label='Iteration Time (ms)')
                plt.xlabel('Training Step')
                plt.ylabel('Time (ms)')
                plt.title(f'{model_name} Iteration Time During Training')
                plt.grid(True)
                
                # Add throughput as second axis
                ax2 = plt.gca().twinx()
                ax2.plot(steps[window-1:], throughput_smooth, 'r-', label='Throughput')
                ax2.set_ylabel('Samples/second', color='r')
                ax2.tick_params(axis='y', labelcolor='r')
                
                # Add both legends
                lines1, labels1 = plt.gca().get_legend_handles_labels()
                lines2, labels2 = ax2.get_legend_handles_labels()
                ax2.legend(lines1 + lines2, labels1 + labels2, loc='upper right')
            
            plt.tight_layout()
            plt.savefig(os.path.join(self.log_dir, f"{model_name}_training_profile.png"))
            plt.close()


class FlopsCalculator:
    """
    Utility for calculating FLOPs (Floating Point Operations) for neural networks
    
    This class provides methods for estimating the computational complexity
    of neural network models in terms of FLOPs.
    
    Args:
        model (torch.nn.Module): PyTorch model
        input_shape (Tuple): Input shape for the model (including batch dimension)
    """
    
    def __init__(self, model: torch.nn.Module, input_shape: Tuple):
        self.model = model
        self.input_shape = input_shape
        self.flops_per_module = {}
        self.total_flops = 0
    
    def calculate_flops(self) -> Dict[str, Union[int, Dict[str, int]]]:
        """
        Calculate FLOPs for the model
        
        Returns:
            Dict[str, Union[int, Dict[str, int]]]: FLOPs statistics
        """
        # Create input tensor
        input_tensor = torch.randn(self.input_shape)
        
        # Dictionary to track tensor shapes
        tensor_shapes = {}
        
        # Register hooks to capture input and output shapes
        handles = []
        
        def pre_hook(name):
            def hook(module, input):
                if input:
                    tensor_shapes[name + "_input"] = [tuple(x.shape) for x in input if isinstance(x, torch.Tensor)]
                return None
            return hook
        
        def post_hook(name):
            def hook(module, input, output):
                if isinstance(output, torch.Tensor):
                    tensor_shapes[name + "_output"] = tuple(output.shape)
                elif isinstance(output, tuple) and all(isinstance(x, torch.Tensor) for x in output):
                    tensor_shapes[name + "_output"] = [tuple(x.shape) for x in output]
                return None
            return hook
        
        # Register hooks
        for name, module in self.model.named_modules():
            if name:  # Skip the root module
                handles.append(module.register_forward_pre_hook(pre_hook(name)))
                handles.append(module.register_forward_hook(post_hook(name)))
        
        # Forward pass to capture shapes
        with torch.no_grad():
            _ = self.model(input_tensor)
        
        # Remove hooks
        for handle in handles:
            handle.remove()
        
        # Calculate FLOPs for each module
        for name, module in self.model.named_modules():
            if name:
                module_flops = self._calculate_module_flops(module, name, tensor_shapes)
                if module_flops > 0:
                    self.flops_per_module[name] = module_flops
                    self.total_flops += module_flops
        
        # Sort modules by FLOPs
        sorted_modules = sorted(
            self.flops_per_module.items(), 
            key=lambda x: x[1], 
            reverse=True
        )
        
        # Compile results
        results = {
            "total_flops": self.total_flops,
            "total_gflops": self.total_flops / (10**9),
            "module_flops": dict(sorted_modules),
        }
        
        # Log summary
        logger.info(f"FLOPs Analysis:")
        logger.info(f"  Total FLOPs: {results['total_flops']:,}")
        logger.info(f"  Total GFLOPs: {results['total_gflops']:.4f}")
        logger.info(f"  Top 5 modules by FLOPs:")
        for i, (name, flops) in enumerate(sorted_modules[:5]):
            logger.info(f"    {i+1}. {name}: {flops:,} ({flops/self.total_flops:.2%})")
        
        return results
    
    def _calculate_module_flops(
        self, 
        module: torch.nn.Module, 
        name: str, 
        tensor_shapes: Dict[str, Any]
    ) -> int:
        """
        Calculate FLOPs for a specific module
        
        Args:
            module (torch.nn.Module): Module to analyze
            name (str): Module name
            tensor_shapes (Dict[str, Any]): Dictionary of tensor shapes
            
        Returns:
            int: Estimated FLOPs for the module
        """
        if not (name + "_input" in tensor_shapes and name + "_output" in tensor_shapes):
            return 0
        
        input_shapes = tensor_shapes[name + "_input"]
        output_shape = tensor_shapes[name + "_output"]
        
        # Handle different module types
        if isinstance(module, torch.nn.Conv2d):
            # FLOPs for Conv2d: (K_h * K_w * C_in * C_out * H_out * W_out) / groups
            if not input_shapes:
                return 0
                
            input_shape = input_shapes[0]
            if len(input_shape) != 4:
                return 0
                
            batch_size, in_channels, in_h, in_w = input_shape
            
            if not isinstance(output_shape, tuple) or len(output_shape) != 4:
                return 0
                
            out_channels = output_shape[1]
            out_h, out_w = output_shape[2], output_shape[3]
            
            kernel_h, kernel_w = module.kernel_size
            groups = module.groups
            
            # Each output element requires (kernel_h * kernel_w * in_channels / groups) multiplications
            # and (kernel_h * kernel_w * in_channels / groups - 1) additions
            flops_per_element = 2 * kernel_h * kernel_w * in_channels * out_channels / groups
            total_flops = int(batch_size * out_h * out_w * flops_per_element)
            
            # Add bias if present
            if module.bias is not None:
                total_flops += batch_size * out_h * out_w * out_channels
            
            return total_flops
            
        elif isinstance(module, torch.nn.Linear):
            # FLOPs for Linear: 2 * in_features * out_features * batch_size
            if not input_shapes:
                return 0
                
            input_shape = input_shapes[0]
            if len(input_shape) < 2:
                return 0
                
            batch_size = input_shape[0]
            in_features = module.in_features
            out_features = module.out_features
            
            # Each output element requires in_features multiplications and in_features-1 additions
            flops_per_element = 2 * in_features * out_features
            total_flops = int(batch_size * flops_per_element)
            
            # Add bias if present
            if module.bias is not None:
                total_flops += batch_size * out_features
            
            return total_flops
            
        elif isinstance(module, torch.nn.BatchNorm2d):
            # FLOPs for BatchNorm2d: 2 * batch_size * channels * height * width
            if not input_shapes:
                return 0
                
            input_shape = input_shapes[0]
            if len(input_shape) != 4:
                return 0
                
            batch_size, channels, height, width = input_shape
            
            # Each element requires 2 operations (normalization and scale/shift)
            return int(2 * batch_size * channels * height * width)
            
        elif isinstance(module, torch.nn.ReLU) or isinstance(module, torch.nn.LeakyReLU):
            # FLOPs for ReLU: batch_size * elements (1 comparison per element)
            if not input_shapes:
                return 0
                
            input_shape = input_shapes[0]
            elements = 1
            for dim in input_shape:
                elements *= dim
            
            return int(elements)
            
        elif isinstance(module, torch.nn.MaxPool2d) or isinstance(module, torch.nn.AvgPool2d):
            # FLOPs for Pooling: batch_size * output_elements * kernel_size^2
            if not input_shapes:
                return 0
                
            input_shape = input_shapes[0]
            if len(input_shape) != 4:
                return 0
                
            batch_size = input_shape[0]
            channels = input_shape[1]
            
            if not isinstance(output_shape, tuple) or len(output_shape) != 4:
                return 0
                
            out_h, out_w = output_shape[2], output_shape[3]
            
            if isinstance(module.kernel_size, int):
                kernel_size = module.kernel_size ** 2
            else:
                kernel_size = module.kernel_size[0] * module.kernel_size[1]
            
            # Each output element requires kernel_size comparisons (MaxPool) or additions (AvgPool)
            return int(batch_size * channels * out_h * out_w * kernel_size)
        
        # Default: return 0 for unsupported module types
        return 0
    
    def generate_flops_report(self) -> str:
        """
        Generate detailed FLOPs report
        
        Returns:
            str: Path to saved report
        """
        # Calculate FLOPs
        flops_data = self.calculate_flops()
        
        # Add model parameters count
        total_params = sum(p.numel() for p in self.model.parameters())
        trainable_params = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
        
        flops_data["total_params"] = total_params
        flops_data["trainable_params"] = trainable_params
        
        # Calculate FLOPs per parameter
        if total_params > 0:
            flops_data["flops_per_param"] = self.total_flops / total_params
        
        # Save report
        report_path = os.path.join(
            "./flops_reports", 
            f"flops_{self.model.__class__.__name__}_{time.strftime('%Y%m%d_%H%M%S')}.json"
        )
        
        os.makedirs(os.path.dirname(report_path), exist_ok=True)
        
        with open(report_path, "w") as f:
            json.dump(flops_data, f, indent=2)
        
        logger.info(f"FLOPs report saved to {report_path}")
        
        # Generate visualization
        self._generate_flops_visualization(flops_data, report_path.replace(".json", ".png"))
        
        return report_path
    
    def _generate_flops_visualization(self, flops_data: Dict[str, Any], output_path: str) -> None:
        """
        Generate visualization of FLOPs distribution
        
        Args:
            flops_data (Dict[str, Any]): FLOPs data
            output_path (str): Path to save visualization
        """
        # Get top 10 modules by FLOPs
        module_flops = flops_data["module_flops"]
        sorted_modules = sorted(
            module_flops.items(), 
            key=lambda x: x[1], 
            reverse=True
        )
        
        top_n = 10
        top_modules = sorted_modules[:top_n]
        
        # Calculate other modules' total
        other_flops = sum(flops for _, flops in sorted_modules[top_n:])
        
        # Prepare data for pie chart
        if other_flops > 0:
            names = [name.split(".")[-1] for name, _ in top_modules] + ["Other"]
            values = [flops for _, flops in top_modules] + [other_flops]
        else:
            names = [name.split(".")[-1] for name, _ in top_modules]
            values = [flops for _, flops in top_modules]
        
        # Create figure
        plt.figure(figsize=(10, 12))
        
        # Create pie chart
        plt.subplot(2, 1, 1)
        explode = [0.1 if i == 0 else 0 for i in range(len(names))]
        plt.pie(
            values, 
            explode=explode,
            labels=names, 
            autopct='%1.1f%%',
            shadow=True, 
            startangle=90
        )
        plt.axis('equal')
        plt.title(f'FLOPs Distribution ({flops_data["total_gflops"]:.2f} GFLOPs)')
        
        # Create bar chart
        plt.subplot(2, 1, 2)
        y_pos = range(len(names))
        plt.barh(y_pos, [v / 10**9 for v in values], align='center')
        plt.yticks(y_pos, names)
        plt.xlabel('GFLOPs')
        plt.title('Module Computational Complexity')
        
        # Add text with model details
        model_info = (
            f"Model: {self.model.__class__.__name__}\n"
            f"Total FLOPs: {flops_data['total_flops']:,}\n"
            f"Total GFLOPs: {flops_data['total_gflops']:.4f}\n"
            f"Total Parameters: {flops_data['total_params']:,}\n"
            f"Trainable Parameters: {flops_data['trainable_params']:,}\n"
        )
        
        plt.figtext(0.5, 0.01, model_info, ha='center', fontsize=10, bbox={"facecolor": "orange", "alpha": 0.2})
        
        plt.tight_layout(rect=[0, 0.05, 1, 0.95])
        plt.savefig(output_path)
        plt.close()



"""
Evolutionary architecture search using PyGAD
"""
import logging
import numpy as np
from typing import List, Dict, Any, Callable, Optional, Union, Tuple

import pygad
import torch
from torch.utils.data import DataLoader

from ..optimization.optuna_optimizer import OptunaOptimizer

logger = logging.getLogger(__name__)


class ArchitectureEncoding:
    """
    Base class for encoding and decoding neural network architectures
    
    This abstract class defines the interface for architecture encoding and decoding.
    Subclasses should implement encode() and decode() methods.
    """
    
    def encode(self, architecture: Any) -> np.ndarray:
        """
        Encode architecture into a genetic representation
        
        Args:
            architecture: Architecture to encode
            
        Returns:
            np.ndarray: Genetic representation
        """
        raise NotImplementedError("Subclasses must implement encode()")
    
    def decode(self, genes: np.ndarray) -> Any:
        """
        Decode genetic representation into an architecture
        
        Args:
            genes (np.ndarray): Genetic representation
            
        Returns:
            Any: Decoded architecture
        """
        raise NotImplementedError("Subclasses must implement decode()")


class MLPArchitectureEncoding(ArchitectureEncoding):
    """
    Encoding for multilayer perceptron architectures
    
    This class encodes and decodes MLP architectures using binary representation.
    
    Args:
        input_size (int): Input size
        output_size (int): Output size
        max_layers (int, optional): Maximum number of hidden layers. Defaults to 5.
        min_neurons (int, optional): Minimum neurons per layer. Defaults to 16.
        max_neurons (int, optional): Maximum neurons per layer. Defaults to 512.
        bits_per_layer (int, optional): Bits used to encode each layer. Defaults to 8.
    """
    
    def __init__(
        self,
        input_size: int,
        output_size: int,
        max_layers: int = 5,
        min_neurons: int = 16,
        max_neurons: int = 512,
        bits_per_layer: int = 8
    ):
        self.input_size = input_size
        self.output_size = output_size
        self.max_layers = max_layers
        self.min_neurons = min_neurons
        self.max_neurons = max_neurons
        self.bits_per_layer = bits_per_layer
        
        # Total number of genes
        self.num_genes = max_layers * bits_per_layer
    
    def encode(self, architecture: List[int]) -> np.ndarray:
        """
        Encode architecture into a genetic representation
        
        Args:
            architecture (List[int]): List of layer sizes
            
        Returns:
            np.ndarray: Binary encoded architecture
        """
        # Remove input and output layer from the encoding
        hidden_layers = architecture[1:-1]
        
        # Initialize genes
        genes = np.zeros(self.num_genes, dtype=int)
        
        # Encode each hidden layer
        for i, neurons in enumerate(hidden_layers):
            if i >= self.max_layers:
                break
                
            # Scale neurons to 0-255 range
            scaled = int((neurons - self.min_neurons) / 
                         (self.max_neurons - self.min_neurons) * (2**self.bits_per_layer - 1))
            
            # Convert to binary
            binary = [(scaled >> bit) & 1 for bit in range(self.bits_per_layer)]
            
            # Set genes
            start_idx = i * self.bits_per_layer
            genes[start_idx:start_idx + self.bits_per_layer] = binary
        
        return genes
    
    def decode(self, genes: np.ndarray) -> List[int]:
        """
        Decode genetic representation into an architecture
        
        Args:
            genes (np.ndarray): Binary encoded architecture
            
        Returns:
            List[int]: Layer sizes including input and output
        """
        architecture = [self.input_size]
        
        # Decode each hidden layer
        for i in range(self.max_layers):
            start_idx = i * self.bits_per_layer
            binary = genes[start_idx:start_idx + self.bits_per_layer]
            
            # Convert binary to decimal
            decimal = sum(bit * (2**idx) for idx, bit in enumerate(binary))
            
            # Scale to neuron range
            neurons = int(self.min_neurons + (decimal / (2**self.bits_per_layer - 1)) * 
                          (self.max_neurons - self.min_neurons))
            
            # Skip layers with too few neurons (allows variable depth networks)
            if neurons >= self.min_neurons:
                architecture.append(neurons)
        
        architecture.append(self.output_size)
        return architecture


class CNNArchitectureEncoding(ArchitectureEncoding):
    """
    Encoding for convolutional neural network architectures
    
    This class encodes and decodes CNN architectures using binary representation.
    
    Args:
        input_shape (Tuple[int, int, int]): Input shape (channels, height, width)
        output_size (int): Output size
        max_conv_layers (int, optional): Maximum number of conv layers. Defaults to 5.
        max_fc_layers (int, optional): Maximum number of fully connected layers. Defaults to 3.
        min_filters (int, optional): Minimum filters per layer. Defaults to 8.
        max_filters (int, optional): Maximum filters per layer. Defaults to 256.
        min_fc_neurons (int, optional): Minimum neurons per FC layer. Defaults to 16.
        max_fc_neurons (int, optional): Maximum neurons per FC layer. Defaults to 1024.
        bits_per_conv_layer (int, optional): Bits for conv layer. Defaults to 12.
        bits_per_fc_layer (int, optional): Bits for FC layer. Defaults to 8.
    """
    
    def __init__(
        self,
        input_shape: Tuple[int, int, int],
        output_size: int,
        max_conv_layers: int = 5,
        max_fc_layers: int = 3,
        min_filters: int = 8,
        max_filters: int = 256,
        min_fc_neurons: int = 16,
        max_fc_neurons: int = 1024,
        bits_per_conv_layer: int = 12,
        bits_per_fc_layer: int = 8
    ):
        self.input_shape = input_shape
        self.output_size = output_size
        self.max_conv_layers = max_conv_layers
        self.max_fc_layers = max_fc_layers
        self.min_filters = min_filters
        self.max_filters = max_filters
        self.min_fc_neurons = min_fc_neurons
        self.max_fc_neurons = max_fc_neurons
        self.bits_per_conv_layer = bits_per_conv_layer
        self.bits_per_fc_layer = bits_per_fc_layer
        
        # Bit allocation for each conv layer:
        # - 8 bits for filter count
        # - 2 bits for kernel size (options: 3, 5, 7, 9)
        # - 1 bit for pooling (yes/no)
        # - 1 bit for layer activation (yes/no)
        
        # Total number of genes
        self.num_genes = (max_conv_layers * bits_per_conv_layer + 
                         max_fc_layers * bits_per_fc_layer)
    
    def decode(self, genes: np.ndarray) -> Tuple[List[Dict[str, Any]], List[int]]:
        """
        Decode genetic representation into a CNN architecture
        
        Args:
            genes (np.ndarray): Binary encoded architecture
            
        Returns:
            Tuple[List[Dict], List[int]]: Convolutional layers and FC layer sizes
        """
        # Decode convolutional layers
        conv_layers = []
        channels, height, width = self.input_shape
        
        for i in range(self.max_conv_layers):
            start_idx = i * self.bits_per_conv_layer
            layer_genes = genes[start_idx:start_idx + self.bits_per_conv_layer]
            
            # Extract filter count (8 bits)
            filter_bits = layer_genes[:8]
            filter_decimal = sum(bit * (2**idx) for idx, bit in enumerate(filter_bits))
            filters = int(self.min_filters + (filter_decimal / (2**8 - 1)) * 
                         (self.max_filters - self.min_filters))
            
            # Skip if too few filters (allows variable depth)
            if filters < self.min_filters:
                continue
            
            # Extract kernel size (2 bits)
            kernel_bits = layer_genes[8:10]
            kernel_decimal = sum(bit * (2**idx) for idx, bit in enumerate(kernel_bits))
            kernel_options = [3, 5, 7, 9]
            kernel_size = kernel_options[kernel_decimal % len(kernel_options)]
            
            # Extract pooling (1 bit)
            pool_bit = layer_genes[10]
            pool_size = 2 if pool_bit else 1
            
            # Create layer config
            layer = {
                'filters': filters,
                'kernel_size': kernel_size,
                'pool_size': pool_size,
                'padding': kernel_size // 2  # Same padding
            }
            
            # Add to layers
            conv_layers.append(layer)
            
            # Update dimensions for next layer
            height = height // pool_size
            width = width // pool_size
            channels = filters
            
            # Break if dimensions become too small
            if height < 2 or width < 2:
                break
        
        # Decode fully connected layers
        fc_layers = []
        start_idx = self.max_conv_layers * self.bits_per_conv_layer
        
        for i in range(self.max_fc_layers):
            layer_start = start_idx + i * self.bits_per_fc_layer
            layer_genes = genes[layer_start:layer_start + self.bits_per_fc_layer]
            
            # Convert binary to decimal
            decimal = sum(bit * (2**idx) for idx, bit in enumerate(layer_genes))
            
            # Scale to neuron range
            neurons = int(self.min_fc_neurons + (decimal / (2**self.bits_per_fc_layer - 1)) * 
                         (self.max_fc_neurons - self.min_fc_neurons))
            
            # Skip layers with too few neurons (allows variable depth)
            if neurons >= self.min_fc_neurons:
                fc_layers.append(neurons)
        
        fc_layers.append(self.output_size)
        
        return conv_layers, fc_layers
    
    def encode(self, architecture: Tuple[List[Dict[str, Any]], List[int]]) -> np.ndarray:
        """
        Encode CNN architecture into a genetic representation
        
        Args:
            architecture: Tuple of conv layers and FC layer sizes
            
        Returns:
            np.ndarray: Binary encoded architecture
        """
        conv_layers, fc_layers = architecture
        
        # Initialize genes
        genes = np.zeros(self.num_genes, dtype=int)
        
        # Encode convolutional layers
        for i, layer in enumerate(conv_layers):
            if i >= self.max_conv_layers:
                break
                
            start_idx = i * self.bits_per_conv_layer
            
            # Encode filter count (8 bits)
            filters = layer.get('filters', self.min_filters)
            filter_scaled = int((filters - self.min_filters) / 
                              (self.max_filters - self.min_filters) * (2**8 - 1))
            filter_binary = [(filter_scaled >> bit) & 1 for bit in range(8)]
            genes[start_idx:start_idx + 8] = filter_binary
            
            # Encode kernel size (2 bits)
            kernel_size = layer.get('kernel_size', 3)
            kernel_options = [3, 5, 7, 9]
            kernel_idx = kernel_options.index(kernel_size) if kernel_size in kernel_options else 0
            kernel_binary = [(kernel_idx >> bit) & 1 for bit in range(2)]
            genes[start_idx + 8:start_idx + 10] = kernel_binary
            
            # Encode pooling (1 bit)
            pool_size = layer.get('pool_size', 1)
            genes[start_idx + 10] = 1 if pool_size > 1 else 0
        
        # Encode fully connected layers
        fc_start_idx = self.max_conv_layers * self.bits_per_conv_layer
        
        for i, neurons in enumerate(fc_layers[:-1]):  # Skip output layer
            if i >= self.max_fc_layers:
                break
                
            start_idx = fc_start_idx + i * self.bits_per_fc_layer
            
            # Scale neurons to binary range
            scaled = int((neurons - self.min_fc_neurons) / 
                         (self.max_fc_neurons - self.min_fc_neurons) * (2**self.bits_per_fc_layer - 1))
            
            # Convert to binary
            binary = [(scaled >> bit) & 1 for bit in range(self.bits_per_fc_layer)]
            
            # Set genes
            genes[start_idx:start_idx + self.bits_per_fc_layer] = binary
        
        return genes


class EvolutionarySearch:
    """
    Evolutionary architecture search using PyGAD
    
    This class performs evolutionary search for neural network architectures
    using genetic algorithms.
    
    Args:
        encoding (ArchitectureEncoding): Architecture encoding/decoding
        fitness_function (Callable): Function to evaluate architecture fitness
        population_size (int, optional): Size of population. Defaults to 20.
        num_generations (int, optional): Number of generations. Defaults to 20.
        crossover_prob (float, optional): Crossover probability. Defaults to 0.7.
        mutation_prob (float, optional): Mutation probability. Defaults to 0.3.
        parent_selection (str, optional): Parent selection method. Defaults to "sss".
        K_tournament (int, optional): Tournament size if using tournament selection. Defaults to 3.
    """
    
    def __init__(
        self,
        encoding: ArchitectureEncoding,
        fitness_function: Callable[[Any], float],
        population_size: int = 20,
        num_generations: int = 20,
        crossover_prob: float = 0.7,
        mutation_prob: float = 0.3,
        parent_selection: str = "sss",
        K_tournament: int = 3
    ):
        self.encoding = encoding
        self.fitness_function = fitness_function
        self.population_size = population_size
        self.num_generations = num_generations
        self.crossover_prob = crossover_prob
        self.mutation_prob = mutation_prob
        self.parent_selection = parent_selection
        self.K_tournament = K_tournament
        
        self.ga_instance = None
        self.best_solution = None
        self.best_fitness = None
    
    def _fitness_wrapper(self, solution, solution_idx):
        """
        Wrapper for the fitness function
        
        Args:
            solution: Genetic representation
            solution_idx: Index in population
            
        Returns:
            float: Fitness value
        """
        # Decode the solution
        architecture = self.encoding.decode(solution)
        
        try:
            # Evaluate the architecture
            fitness = self.fitness_function(architecture)
            return fitness
        except Exception as e:
            logger.error(f"Error evaluating architecture: {e}")
            return -1e9  # Very low fitness for failed architectures
    
    def evolve(self, initial_population=None):
        """
        Run the evolutionary search
        
        Args:
            initial_population: Optional initial population
            
        Returns:
            Tuple: Best architecture, best fitness
        """
        # Initialize GA instance
        self.ga_instance = pygad.GA(
            num_generations=self.num_generations,
            num_parents_mating=max(2, self.population_size // 5),
            sol_per_pop=self.population_size,
            num_genes=self.encoding.num_genes,
            fitness_func=self._fitness_wrapper,
            gene_type=int,
            gene_space=[0, 1],  # Binary genes
            crossover_probability=self.crossover_prob,
            mutation_probability=self.mutation_prob,
            parent_selection_type=self.parent_selection,
            K_tournament=self.K_tournament,
            initial_population=initial_population,
            save_best_solutions=True,
            stop_criteria=["reach_0.9999", "saturate_10"]
        )
        
        # Run GA
        self.ga_instance.run()
        
        # Get best solution
        solution, fitness, _ = self.ga_instance.best_solution()
        
        # Decode best architecture
        best_architecture = self.encoding.decode(solution)
        
        self.best_solution = solution
        self.best_fitness = fitness
        
        logger.info(f"Best fitness: {fitness}")
        logger.info(f"Best architecture: {best_architecture}")
        
        return best_architecture, fitness
    
    def get_solution_summary(self):
        """
        Get summary of the search
        
        Returns:
            Dict: Summary information
        """
        if self.ga_instance is None:
            raise ValueError("No search has been performed yet. Call evolve() first.")
        
        return {
            "best_solution": self.best_solution.tolist() if self.best_solution is not None else None,
            "best_fitness": self.best_fitness,
            "best_architecture": self.encoding.decode(self.best_solution) if self.best_solution is not None else None,
            "num_generations": self.ga_instance.generations_completed,
            "fitness_history": [gen.mean() for gen in self.ga_instance.best_solutions_fitness]
        }


class MultilevelOptimization:
    """
    Multi-level optimization combining evolutionary search and hyperparameter optimization
    
    This class integrates PyGAD for architecture search and Optuna for hyperparameter
    optimization in a multi-level approach.
    
    Args:
        model_class (type): Lightning model class
        encoding (ArchitectureEncoding): Architecture encoding/decoding
        train_loader (DataLoader): Training data loader
        val_loader (DataLoader): Validation data loader
        test_loader (Optional[DataLoader], optional): Test data loader. Defaults to None.
        metric_name (str, optional): Metric to optimize. Defaults to "val_acc".
        direction (str, optional): Optimization direction. Defaults to "maximize".
        fixed_model_kwargs (Dict[str, Any], optional): Fixed model parameters. Defaults to {}.
        hp_param_generators (Dict[str, Callable], optional): Hyperparameter generators. Defaults to {}.
        population_size (int, optional): GA population size. Defaults to 10.
        num_generations (int, optional): GA generations. Defaults to 10.
        n_trials (int, optional): Optuna trials per architecture. Defaults to 5.
        max_epochs (int, optional): Max training epochs. Defaults to 10.
    """
    
    def __init__(
        self,
        model_class: type,
        encoding: ArchitectureEncoding,
        train_loader: DataLoader,
        val_loader: DataLoader,
        test_loader: Optional[DataLoader] = None,
        metric_name: str = "val_acc",
        direction: str = "maximize",
        fixed_model_kwargs: Dict[str, Any] = None,
        hp_param_generators: Dict[str, Callable] = None,
        population_size: int = 10,
        num_generations: int = 10,
        n_trials: int = 5,
        max_epochs: int = 10
    ):
        self.model_class = model_class
        self.encoding = encoding
        self.train_loader = train_loader
        self.val_loader = val_loader
        self.test_loader = test_loader
        self.metric_name = metric_name
        self.direction = direction
        self.fixed_model_kwargs = fixed_model_kwargs or {}
        self.hp_param_generators = hp_param_generators or {}
        self.population_size = population_size
        self.num_generations = num_generations
        self.n_trials = n_trials
        self.max_epochs = max_epochs
        
        # Results tracking
        self.best_architecture = None
        self.best_hyperparams = None
        self.best_fitness = None if direction == "maximize" else float('inf')
        self.architecture_fitness_history = []
    
    def optimize_hyperparams(self, architecture):
        """
        Optimize hyperparameters for a given architecture
        
        Args:
            architecture: Neural network architecture
            
        Returns:
            Tuple[Dict[str, Any], float]: Best hyperparameters and fitness
        """
        # For MLPArchitectureEncoding
        if isinstance(architecture, list):
            model_kwargs = {
                **self.fixed_model_kwargs,
                'architecture': architecture
            }
        # For CNNArchitectureEncoding
        elif isinstance(architecture, tuple) and len(architecture) == 2:
            conv_architecture, fc_architecture = architecture
            model_kwargs = {
                **self.fixed_model_kwargs,
                'conv_architecture': conv_architecture,
                'fc_architecture': fc_architecture
            }
        else:
            raise ValueError(f"Unsupported architecture format: {type(architecture)}")
        
        # Create optimizer
        optimizer = OptunaOptimizer(
            model_class=self.model_class,
            model_kwargs=model_kwargs,
            train_loader=self.train_loader,
            val_loader=self.val_loader,
            metric_name=self.metric_name,
            direction=self.direction,
            max_epochs=self.max_epochs,
            n_trials=self.n_trials
        )
        
        # Add hyperparameter spaces
        for param_name, generator in self.hp_param_generators.items():
            optimizer.add_param_space(param_name, generator)
        
        # If no hyperparameters were added, add a default learning rate
        if not self.hp_param_generators:
            optimizer.add_float_param('learning_rate', 1e-4, 1e-1, log=True)
        
        # Optimize
        best_params = optimizer.optimize()
        
        # Get best value
        study = optimizer.study
        best_value = study.best_value
        if self.direction == 'maximize':
            best_value = -best_value  # Negate if maximizing
        
        return best_params, best_value
    
    def fitness_function(self, architecture):
        """
        Fitness function for evolutionary search
        
        Args:
            architecture: Neural network architecture
            
        Returns:
            float: Fitness value (higher is better)
        """
        logger.info(f"Evaluating architecture: {architecture}")
        
        try:
            # Optimize hyperparameters for this architecture
            best_params, best_fitness = self.optimize_hyperparams(architecture)
            
            # For minimization, invert the fitness
            fitness = best_fitness if self.direction == "maximize" else -best_fitness
            
            # Track architecture and fitness
            self.architecture_fitness_history.append((architecture, fitness, best_params))
            
            # Update best overall if better
            if self.direction == "maximize":
                if fitness > self.best_fitness:
                    self.best_architecture = architecture
                    self.best_hyperparams = best_params
                    self.best_fitness = fitness
            else:
                if fitness < self.best_fitness:
                    self.best_architecture = architecture
                    self.best_hyperparams = best_params
                    self.best_fitness = fitness
            
            logger.info(f"Architecture fitness: {fitness}")
            return fitness
        
        except Exception as e:
            logger.error(f"Error in fitness evaluation: {e}")
            return -1e9 if self.direction == "maximize" else 1e9
    
    def search(self):
        """
        Perform multi-level optimization
        
        Returns:
            Tuple: Best architecture, best hyperparameters, best model
        """
        # Create evolutionary search
        evolution = EvolutionarySearch(
            encoding=self.encoding,
            fitness_function=self.fitness_function,
            population_size=self.population_size,
            num_generations=self.num_generations
        )
        
        # Run evolutionary search
        evolution.evolve()
        
        # Train final model with best architecture and hyperparameters
        if self.best_architecture is not None and self.best_hyperparams is not None:
            # For MLPArchitectureEncoding
            if isinstance(self.best_architecture, list):
                model_kwargs = {
                    **self.fixed_model_kwargs,
                    **self.best_hyperparams,
                    'architecture': self.best_architecture
                }
            # For CNNArchitectureEncoding
            elif isinstance(self.best_architecture, tuple) and len(self.best_architecture) == 2:
                conv_architecture, fc_architecture = self.best_architecture
                model_kwargs = {
                    **self.fixed_model_kwargs,
                    **self.best_hyperparams,
                    'conv_architecture': conv_architecture,
                    'fc_architecture': fc_architecture
                }
            
            # Create and train model
            best_model = self.model_class(**model_kwargs)
            
            trainer = L.Trainer(max_epochs=self.max_epochs * 2)
            trainer.fit(best_model, self.train_loader, self.val_loader)
            
            # Test if test loader provided
            if self.test_loader:
                trainer.test(best_model, self.test_loader)
            
            return self.best_architecture, self.best_hyperparams, best_model
        
        return None, None, None
    
    def get_search_summary(self):
        """
        Get summary of the search
        
        Returns:
            Dict: Summary information
        """
        return {
            "best_architecture": self.best_architecture,
            "best_hyperparams": self.best_hyperparams,
            "best_fitness": self.best_fitness,
            "architecture_history": self.architecture_fitness_history
        }



"""
Example usage for CIFAR-10 classification with PyDeevo using CNNs
"""
import torch
import torch.nn.functional as F
import torchvision
from torchvision import transforms
from torch.utils.data import DataLoader
import lightning as L

from pydeevo import PyDeevo
from pydeevo.models.base import CNNModule


def main():
    """Main function for CIFAR-10 example"""
    # Set up PyDeevo
    pydeevo = PyDeevo(base_dir="./cifar10_example")
    
    # Load CIFAR-10 dataset
    transform_train = transforms.Compose([
        transforms.RandomCrop(32, padding=4),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
    ])
    
    transform_test = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
    ])
    
    train_dataset = torchvision.datasets.CIFAR10('./data', train=True, download=True, transform=transform_train)
    test_dataset = torchvision.datasets.CIFAR10('./data', train=False, transform=transform_test)
    
    # Split training dataset into train and validation
    train_size = int(0.8 * len(train_dataset))
    val_size = len(train_dataset) - train_size
    train_dataset, val_dataset = torch.utils.data.random_split(train_dataset, [train_size, val_size])
    
    # Create data loaders
    train_loader = DataLoader(train_dataset, batch_size=64, shuffle=True)
    val_loader = DataLoader(val_dataset, batch_size=64)
    test_loader = DataLoader(test_dataset, batch_size=64)
    
    # Define input shape and output size for CIFAR-10
    input_shape = (3, 32, 32)  # CIFAR-10 images (channels, height, width)
    output_size = 10  # 10 classes
    
    print("Starting architecture search...")
    
    # Evolve architecture
    result = pydeevo.evolve_architecture(
        input_shape=input_shape,
        output_size=output_size,
        train_loader=train_loader,
        val_loader=val_loader,
        network_type="cnn",
        population_size=5,  # Small population for example
        num_generations=3,  # Few generations for example
        hp_trials_per_arch=2,  # Few trials for example
        max_epochs=5  # Few epochs for example
    )
    
    best_architecture = result["best_architecture"]
    best_hyperparams = result["best_hyperparameters"]
    
    conv_architecture, fc_architecture = best_architecture
    
    print(f"Best convolutional layers: {conv_architecture}")
    print(f"Best fully connected layers: {fc_architecture}")
    print(f"Best hyperparameters: {best_hyperparams}")
    
    # Create a model with the best architecture and hyperparameters
    model = CNNModule(
        conv_architecture=conv_architecture,
        fc_architecture=fc_architecture,
        input_shape=input_shape,
        **best_hyperparams
    )
    
    print("Training final model...")
    
    # Train the model
    trainer, metrics = pydeevo.train_model(
        model=model,
        train_loader=train_loader,
        val_loader=val_loader,
        test_loader=test_loader,
        max_epochs=20
    )
    
    print(f"Test metrics: {metrics}")


if __name__ == "__main__":
    main()



"""
Main integration module for PyDeevo
"""
import os
import logging
from typing import Dict, List, Any, Optional, Union, Tuple, Callable

import torch
from torch.utils.data import DataLoader
import optuna
import lightning as L

from ..models.base import FlexibleModule, CNNModule
from ..optimization.optuna_optimizer import OptunaOptimizer
from ..evolution.architecture_search import (
    MLPArchitectureEncoding,
    CNNArchitectureEncoding,
    EvolutionarySearch,
    MultilevelOptimization
)
from ..training.trainer import LightningTrainer, DataModule
from ..utils.helpers import setup_logging, save_dict_to_json, visualize_architecture_comparison

logger = logging.getLogger(__name__)


class PyDeevo:
    """
    Main class integrating all PyDeevo components
    
    This class provides a high-level interface to the PyDeevo framework,
    combining PyTorch Lightning, Optuna, and PyGAD.
    
    Args:
        base_dir (str, optional): Base directory for outputs. Defaults to "./pydeevo_output".
        log_level (int, optional): Logging level. Defaults to logging.INFO.
    """
    
    def __init__(
        self,
        base_dir: str = "./pydeevo_output",
        log_level: int = logging.INFO
    ):
        self.base_dir = base_dir
        
        # Create directories
        os.makedirs(base_dir, exist_ok=True)
        os.makedirs(os.path.join(base_dir, "logs"), exist_ok=True)
        os.makedirs(os.path.join(base_dir, "checkpoints"), exist_ok=True)
        os.makedirs(os.path.join(base_dir, "results"), exist_ok=True)
        
        # Setup logging
        self.logger = setup_logging(
            log_dir=os.path.join(base_dir, "logs"),
            log_level=log_level
        )
    
    def optimize_hyperparameters(
        self,
        model_class: type,
        model_kwargs: Dict[str, Any],
        train_loader: DataLoader,
        val_loader: DataLoader,
        param_spaces: Dict[str, Callable],
        metric_name: str = "val_loss",
        direction: str = "minimize",
        n_trials: int = 20,
        max_epochs: int = 10
    ) -> Dict[str, Any]:
        """
        Optimize hyperparameters using Optuna
        
        Args:
            model_class (type): Lightning model class
            model_kwargs (Dict[str, Any]): Model initialization arguments
            train_loader (DataLoader): Training data loader
            val_loader (DataLoader): Validation data loader
            param_spaces (Dict[str, Callable]): Parameter space generators
            metric_name (str, optional): Metric to optimize. Defaults to "val_loss".
            direction (str, optional): Optimization direction. Defaults to "minimize".
            n_trials (int, optional): Number of trials. Defaults to 20.
            max_epochs (int, optional): Maximum epochs per trial. Defaults to 10.
            
        Returns:
            Dict[str, Any]: Best hyperparameters
        """
        # Create optimizer
        optimizer = OptunaOptimizer(
            model_class=model_class,
            model_kwargs=model_kwargs,
            train_loader=train_loader,
            val_loader=val_loader,
            metric_name=metric_name,
            direction=direction,
            max_epochs=max_epochs,
            n_trials=n_trials,
            study_name=f"{model_class.__name__}_study",
            storage=f"sqlite:///{os.path.join(self.base_dir, 'optuna.db')}"
        )
        
        # Add parameter spaces
        for param_name, generator in param_spaces.items():
            optimizer.add_param_space(param_name, generator)
        
        # Run optimization
        best_params = optimizer.optimize()
        
        # Save results
        results = {
            "model_class": model_class.__name__,
            "best_params": best_params,
            "best_value": optimizer.study.best_value,
            "n_trials": n_trials
        }
        
        save_dict_to_json(
            results,
            os.path.join(self.base_dir, "results", "hyperparameter_optimization.json")
        )
        
        return best_params
    
    def evolve_architecture(
        self,
        input_shape: Union[int, Tuple[int, int, int]],
        output_size: int,
        train_loader: DataLoader,
        val_loader: DataLoader,
        network_type: str = "mlp",
        population_size: int = 10,
        num_generations: int = 10,
        hp_trials_per_arch: int = 5,
        max_epochs: int = 10,
        metric_name: str = "val_acc",
        direction: str = "maximize",
        fixed_model_kwargs: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Evolve neural network architecture using genetic algorithms
        
        Args:
            input_shape (Union[int, Tuple[int, int, int]]): Input shape (scalar for MLP, tuple for CNN)
            output_size (int): Output size
            train_loader (DataLoader): Training data loader
            val_loader (DataLoader): Validation data loader
            network_type (str, optional): Network type ("mlp" or "cnn"). Defaults to "mlp".
            population_size (int, optional): Population size. Defaults to 10.
            num_generations (int, optional): Number of generations. Defaults to 10.
            hp_trials_per_arch (int, optional): Hyperparameter trials per architecture. Defaults to 5.
            max_epochs (int, optional): Maximum epochs per trial. Defaults to 10.
            metric_name (str, optional): Metric to optimize. Defaults to "val_acc".
            direction (str, optional): Optimization direction. Defaults to "maximize".
            fixed_model_kwargs (Optional[Dict[str, Any]], optional): Fixed model arguments. Defaults to None.
            
        Returns:
            Dict[str, Any]: Evolution results
        """
        fixed_model_kwargs = fixed_model_kwargs or {}
        
        # Select model class and encoding based on network type
        if network_type.lower() == "mlp":
            model_class = FlexibleModule
            encoding = MLPArchitectureEncoding(
                input_size=input_shape,
                output_size=output_size,
                max_layers=5
            )
        elif network_type.lower() == "cnn":
            model_class = CNNModule
            encoding = CNNArchitectureEncoding(
                input_shape=input_shape,
                output_size=output_size,
                max_conv_layers=5,
                max_fc_layers=3
            )
        else:
            raise ValueError(f"Unknown network type: {network_type}")
        
        # Define hyperparameter spaces
        hp_param_generators = {
            'learning_rate': lambda trial: trial.suggest_float('learning_rate', 1e-4, 1e-1, log=True),
            'optimizer_kwargs': lambda trial: {
                'weight_decay': trial.suggest_float('weight_decay', 1e-6, 1e-3, log=True)
            }
        }
        
        # Create multi-level optimizer
        optimizer = MultilevelOptimization(
            model_class=model_class,
            encoding=encoding,
            train_loader=train_loader,
            val_loader=val_loader,
            metric_name=metric_name,
            direction=direction,
            fixed_model_kwargs=fixed_model_kwargs,
            hp_param_generators=hp_param_generators,
            population_size=population_size,
            num_generations=num_generations,
            n_trials=hp_trials_per_arch,
            max_epochs=max_epochs
        )
        
        # Run search
        best_architecture, best_hyperparams, best_model = optimizer.search()
        
        # Create result summary
        summary = optimizer.get_search_summary()
        
        # Save best architecture and hyperparameters
        save_dict_to_json(
            summary,
            os.path.join(self.base_dir, "results", "architecture_search.json")
        )
        
        # Visualize architecture comparison
        architectures = [arch for arch, _, _ in summary['architecture_history']]
        fitness_values = [fitness for _, fitness, _ in summary['architecture_history']]
        
        visualize_architecture_comparison(
            architectures,
            fitness_values,
            os.path.join(self.base_dir, "results", "architecture_comparison.png")
        )
        
        # Save model checkpoint if available
        if best_model is not None:
            checkpoint_path = os.path.join(self.base_dir, "checkpoints", "best_evolved_model.ckpt")
            trainer = L.Trainer()
            trainer.save_checkpoint(checkpoint_path)
        
        return {
            "best_architecture": best_architecture,
            "best_hyperparameters": best_hyperparams,
            "summary": summary
        }
    
    def train_model(
        self,
        model: L.LightningModule,
        train_loader: DataLoader,
        val_loader: Optional[DataLoader] = None,
        test_loader: Optional[DataLoader] = None,
        max_epochs: int = 100,
        patience: int = 10,
        monitor: str = "val_loss",
        mode: str = "min"
    ) -> Tuple[L.Trainer, Dict[str, float]]:
        """
        Train a Lightning model
        
        Args:
            model (L.LightningModule): Lightning model
            train_loader (DataLoader): Training data loader
            val_loader (Optional[DataLoader], optional): Validation data loader. Defaults to None.
            test_loader (Optional[DataLoader], optional): Test data loader. Defaults to None.
            max_epochs (int, optional): Maximum epochs. Defaults to 100.
            patience (int, optional): Early stopping patience. Defaults to 10.
            monitor (str, optional): Metric to monitor. Defaults to "val_loss".
            mode (str, optional): Optimization mode. Defaults to "min".
            
        Returns:
            Tuple[L.Trainer, Dict[str, float]]: Trainer and metrics
        """
        # Create trainer
        trainer = LightningTrainer(
            max_epochs=max_epochs,
            patience=patience,
            monitor=monitor,
            mode=mode,
            checkpoint_dir=os.path.join(self.base_dir, "checkpoints"),
            log_dir=os.path.join(self.base_dir, "logs")
        )
        
        # Setup trainer
        trainer.setup_trainer()
        
        # Train model
        trainer.fit(model, train_loader, val_loader)
        
        # Test if test loader provided
        metrics = {}
        if test_loader is not None:
            metrics = trainer.test(model, test_loader)
        
        # Save model summary
        trainer.save_model_summary(model)
        
        return trainer, metrics
    
    def load_best_model(
        self, 
        model_class: type, 
        checkpoint_path: Optional[str] = None, 
        **model_kwargs
    ) -> L.LightningModule:
        """
        Load a model from checkpoint
        
        Args:
            model_class (type): Lightning model class
            checkpoint_path (Optional[str], optional): Checkpoint path. Defaults to None.
            **model_kwargs: Model initialization arguments
            
        Returns:
            L.LightningModule: Loaded model
        """
        if checkpoint_path is None:
            # Try to find the latest checkpoint
            checkpoint_dir = os.path.join(self.base_dir, "checkpoints")
            checkpoints = [f for f in os.listdir(checkpoint_dir) if f.endswith(".ckpt")]
            
            if not checkpoints:
                raise ValueError("No checkpoints found.")
            
            # Use the latest checkpoint
            checkpoint_path = os.path.join(checkpoint_dir, sorted(checkpoints)[-1])
        
        # Load model
        model = model_class.load_from_checkpoint(checkpoint_path, **model_kwargs)
        
        return model
